#!/usr/bin/env python3

"""
Core Tasks

Tasks that are a core part of uPlaybook.
"""

from .internals import (
    Return,
    Exit,
    TemplateStr,
    up_context,
    task,
    Failure,
    RawStr,  # noqa
    CallDepth,
)
from . import internals
from typing import Optional, List, Union, Callable
import pprint
import subprocess
import sys
from types import SimpleNamespace
import argparse
import os
import pwd
import re
import requests


class IgnoreFailure:
    """A context-manager to ignore failures in wrapped tasks.

    Example:

        with core.IgnoreFailures():
            core.run("false")
            if not core.mkdir("/root/failure"):
                print("You are not root")
    """

    def __enter__(self):
        up_context.ignore_failure_count += 1
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        up_context.ignore_failure_count -= 1
        assert up_context.ignore_failure_count >= 0


class Item(dict):
    """
    An (ansible-like) item for processing in a playbook (a file, directory, user...)

    A typical use case of an Item is for looping over many files and setting them
    up, or setting up many filesystem objects in a playbook.  Additionally, if used
    in a "with:" statement, it will place the attributes into the current Jinja2
    namespace.

    This can act as a dictionary (x['attr'] access), SimpleNamespace (x.attr), and also
    is a context manager.  In the context manager case, it forklifts the attributes
    up into the Jinja2 context.

    Examples:

    ```python
    for item in [
            Item(path="foo"),
            Item(path="bar"),
            Item(path="baz"),
            ]:
        fs.builder(path="{{item.dst}}")

    for item in [
            Item(path="foo", action="directory", owner="nobody"),
            Item(path="bar", action="exists"),
            Item(path="/etc/apache2/sites-enabled/foo", notify=restart_apache),
            ]:
        fs.builder(**item)

    with Item(path="foo", action="directory", owner="nobody") as item:
        #  Can access as "dst" as well as "item.dst"
        fs.exists(path="{{dst}}")
        fs.chown(path="{{dst}}", owner="{{owner}}")

    with Item(path="foo", action="directory", owner="nobody"):
        fs.exists(path="{{dst}}")
    ```
    """

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{key}'"
            )

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, key):
        if key in self:
            del self[key]
        else:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{key}'"
            )

    def __enter__(self):
        up_context.item_context.insert(0, self)
        return self

    def __exit__(self, *_):
        up_context.item_context.pop(0)


@task
def debug(msg: Optional[TemplateStr] = None, var: Optional[object] = None) -> Return:
    """
    Display informational message.

    Print a message or pretty-print a variable.

    Args:
        msg: Message to display. (optional, templateable).
        var: Object to pretty-print (optional, templateable).

    Examples:

    ```python
    core.debug(msg="Directory already exists, exiting")
    core.debug(var=ret_value)
    ```
    """

    output = ""
    if msg:
        output = msg + "\n"
    if var:
        output += pprint.pformat(var)
        output = "\n".join("    " + line for line in output.splitlines())
    output = output.rstrip()

    return Return(changed=False, output=output, hide_args=True)


@task
def lookup(var: str) -> object:
    """
    Looks up `var` in the "up context" and returns the value.

    This would typically be similar to Jinja2 rendering "{{var}}", but
    it does not go through "." traversing, so `var` must be a top
    level name.

    Examples:

    ```python
    print(core.lookup('name'))
    ```
    """
    return up_context.get_env()[var]


@task
def render(s: TemplateStr) -> str:
    """
    Render a string as a jinja2 template and return the value

    Args:
        s: Template to render. (templateable).

    Returns:
        Rendered template as a string.

    Examples:

    ```python
    core.render(s="Value of foo: {{foo}}")
    ```
    """
    return s


@task
def run(
    command: TemplateStr,
    shell: bool = True,
    ignore_failure: bool = False,
    change: bool = True,
    creates: Optional[TemplateStr] = None,
) -> object:
    """
    Run a command.  Stdout is returned as `output` in the return object.  Stderr
    and return code are stored in `extra` in return object.

    Args:
        command: Command to run (templateable).
        shell: If False, run `command` without a shell.  Safer.  Default is True:
             allows shell processing of `command` for things like output
             redirection, wildcard expansion, pipelines, etc. (optional, bool)
        ignore_failure: If True, do not treat non-0 return code as a fatal failure.
             This allows testing of return code within playbook.  (optional, bool)
        change: By default, all shell commands are assumed to have caused a change
             to the system and will trigger notifications.  If False, this `command`
             is treated as not changing the system.  (optional, bool)
        creates: If specified, if the path it specifies exists, consider the command
            to have already been run and skip future runes.

    Extra Data:

        stderr: Captured stderr output.
        returncode: The return code of the command.

    Examples:

    ```python
    core.run(command="systemctl restart sshd")
    core.run(command="rm *.foo", shell=False)   #  removes literal file "*.foo"

    r = core.run(command="date", change=False)
    print(f"Current date/time: {{r.output}}")
    print(f"Return code: {{r.extra.returncode}}")

    if core.run(command="grep -q ^user: /etc/passwd", ignore_failure=True, change=False):
        print("User exists")
    ```
    """
    if creates is not None and os.path.exists(creates):
        return Return(changed=False)

    sys.stdout.flush()
    sys.stderr.flush()

    p = subprocess.run(command, shell=shell, text=True, capture_output=True)

    extra = SimpleNamespace()
    extra.stderr = p.stderr
    extra.returncode = p.returncode
    failure = p.returncode != 0

    failure_msg = f"Exit code: {p.returncode}"
    if p.stdout:
        failure_msg += f"\n   >stdout: {p.stdout.rstrip()}"
    if p.stderr:
        failure_msg += f"\n   >stderr: {p.stderr.rstrip()}"

    return Return(
        changed=change,
        failure=failure,
        output=p.stdout.rstrip(),
        extra=extra,
        ignore_failure=ignore_failure,
        raise_exc=Failure(failure_msg) if failure and not ignore_failure else None,
    )


class Argument:
    """
    A playbook argument, this class is used to describe the details of an argument/option.

    Use this to provide details about arguments or options that can be used with a playbook.
    These get turned into command-line arguments that can be described with "--help".
    This is used in combination with the `core.playbook_args()` task.

    Names that include "_" (underscore) will be converted to "-" (hyphen) for the command
    line argument.

    Args:
        name: The name of the argument, this will determine the "--name" of the command-line
            flag and the variable the value is stored in (str).
        description: Detailed information on the argument for use in "--help" output.
            (str, optional)
        type: The type of the argument: str, bool, int, password (default=str)
        default: A default value for the argument.  Arguments without a default
            must be specified in the command-line.  If default is specified, a command-line
            option with "--`name`" will be available.

    Examples:

    ```python
    core.playbook_args(
            core.Argument(name="user"),
            core.Argument(name="hostname", default=None)
            core.Argument(name="enable", type="bool", default=True)
            )
    core.debug(msg="Arguments: user={{ARGS.user}}  hostname={{ARGS.hostname}} enable={{ARGS.enable}}")

    #  Run with "up2 playbookname --hostname=localhost --no-enable username
    ```
    """

    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        type: str = "str",
        default: Optional[object] = None,
    ):
        self.name = name
        self.description = description
        self.type = type
        self.default = default


@task
def playbook_args(
    *options: List[Argument],
) -> None:
    """
    Specify arguments for a playbook.

    Optionally, a playbook may specify that it needs arguments.  If defined,
    this will create an argument parser and command-line arguemnts and
    options.

    Examples:

    ```python
    core.playbook_args(
            core.Argument(name="is_owner", default=False, type="bool"),
            core.Argument(name="user"),
            core.Argument(name="hostname", default="localhost")
            )
    core.debug(msg="Arguments: user={{playbook_args.user}}  hostname={{playbook_args.hostname}}")
    core.debug(msg="Arguments: is_owner={{playbook_args.is_owner}}")
    #  run examples:
    #    up playbook.pb my_username
    #    up playbook.pb --is-owner my_username
    #    up playbook.pb --no-is-owner my_username my_hostname
    ```
    """
    parser = argparse.ArgumentParser(
        prog=f"up:{up_context.playbook_name}", description=up_context.playbook_docstring
    )

    for arg in options:
        arg_type = {
            "bool": bool,
            "str": str,
            "int": int,
            "password": str,
        }[arg.type]

        kw_args: dict[str, object] = {}
        if arg.description is not None:
            kw_args["help"] = arg.description
        if arg.type == "bool":
            kw_args["action"] = argparse.BooleanOptionalAction

        #  name is "--<NAME>" if default is specified, else make it a positional arg
        argument = arg.name if arg.default is None else "--" + arg.name
        if argument.startswith("--"):
            kw_args["dest"] = argument[2:].replace("-", "_")
        else:
            argument = argument.replace("-", "_")

        parser.add_argument(argument, type=arg_type, default=arg.default, **kw_args)

    args, remaining = parser.parse_known_args(up_context.remaining_args)
    up_context.remaining_args = remaining

    #  update up_context.ARGS
    for k, v in vars(args).items():
        setattr(up_context.context["ARGS"], k, v)


@task
def become(user: Union[int, TemplateStr]) -> Return:
    """
    Switch to running as another user in a playbook.

    If used as a context manager, you are switched back to the original user after the context.

    Args:
        user: User name or UID of user to switch to.

    Examples:

    ```python
    core.become(user="nobody")

    with core.become(user="backup"):
        #  to tasks as backup user
        fs.mkfile(path="/tmp/backupfile")
    #  now you are back to the previous user
    ```
    """
    if isinstance(user, str):
        user = pwd.getpwnam(user).pw_uid

    assert isinstance(user, int)
    old_uid = os.getuid()
    os.seteuid(user)

    return Return(changed=False, context_manager=lambda: os.seteuid(old_uid))


@task
def require(user: Union[int, TemplateStr]) -> Return:
    """
    Verify we are running as the specified user.

    Args:
        user: User name or UID of user to verify.  (int or str, templateable)

    Examples:

    ```python
    core.require(user="nobody")
    ```
    """
    if isinstance(user, str):
        user = pwd.getpwnam(user).pw_uid

    assert isinstance(user, int)
    current_uid = os.getuid()

    if current_uid != user:
        Return(
            changed=False,
            failure=True,
            raise_exc=Failure(f"Expected to run as user {user}, got uid={current_uid}"),
        )

    return Return(changed=False)


@task
def fail(msg: TemplateStr) -> Return:
    """
    Abort a playbook run.

    Args:
        msg: Message to display with failure (str, templateable).

    Examples:

    ```python
    core.fail(msg="Unable to download file")
    ```
    """
    return Return(changed=False, failure=True, raise_exc=Failure(msg))


@task
def exit(returncode: int = 0, msg: Union[TemplateStr, str] = "") -> Return:
    """
    End a playbook run.

    Args:
        returncode: Exit code for process, 0 is success, 1-255 are failure (int, default=0).
        msg: Message to display (str, templatable, default "").

    Examples:

    ```python
    core.exit()
    core.exit(returncode=1)
    core.exit(msg="Unable to download file", returncode=1)
    ```
    """
    return Return(
        changed=False, failure=returncode != 0, raise_exc=Exit(msg, returncode)
    )


@task
def notify(handler: Union[Callable, List]) -> Return:
    """
    Add a notify handler to be called later.

    Args:
        handler: A function, or list of functions, that takes no arguments, which is
                called at a later time.  (callable or list)

    Examples:

    ```python
    core.notify(lambda: core.run(command="systemctl restart apache2"))
    core.notify(lambda: fs.remove("tmpdir", recursive=True))
    core.notify([handler1, handler2])
    ```
    """
    if callable(handler):
        up_context.add_handler(handler)
    else:
        for x in handler:
            up_context.add_handler(x)

    return Return(changed=False)


@task
def flush_handlers() -> Return:
    """
    Run any registred handlers.

    Examples:

    ```python
    core.flush_handlers()
    ```
    """
    up_context.flush_handlers()

    return Return(changed=False)


@task
def grep(
    path: TemplateStr,
    search: TemplateStr,
    regex: bool = True,
    ignore_failure: bool = True,
) -> object:
    """
    Look for `search` in the file `path`

    Args:
        path: File location to look for a match in. (templateable)
        search: The string (or regex) to look for. (templateable)
        regex: Do a regex search, if False do a simple string search. (bool, default=True)
        ignore_failure: If True, do not treat file absence as a fatal failure.
             (optional, bool, default=True)

    Examples:

    ```python
    if core.grep(path="/tmp/foo", search="secret=xyzzy"):
        #  code for when the string is found.
    ```
    """
    with open(path, "r") as fp:
        if regex:
            rx = re.compile(search)
            for line in fp.readlines():
                if rx.search(line):
                    return Return(changed=False, failure=False)
        else:
            for line in fp.readlines():
                if search in line:
                    return Return(changed=False, failure=False)

    return Return(
        changed=False,
        failure=True,
        ignore_failure=ignore_failure,
        raise_exc=Failure("No match found") if not ignore_failure else None,
    )


@task
def print(
    msg: TemplateStr,
) -> None:
    """
    uPlaybook print helper, like python print() but does jinja templating.

    Args:
        msg: Message to print. (templateable)

    Examples:

    ```python
    core.print("Arguments: {{playbook_arguments}}")
    ```
    """
    sys.stdout.write(msg + "\n")


@task
def include(playbook: TemplateStr, hoist_vars: bool = True) -> Return:
    """
    Access another playbook from an existing playbook.

    Args:
        playbook: Playbook to include. (templateable)
        hoist_vars: If True (the default), variables set in the playbook are "hoisted"
                up into the main playbook.

    Examples:

    ```python
    core.include(playbook="set_vars.pb")
    ```
    """
    full_playbook_path = os.path.join(up_context.playbook_directory, playbook)
    # pb_info = internals.find_playbook(full_playbook_path)
    # internals.import_script_as_module(
    #    playbook, [pb_info.playbook_file, pb_info.playbook_file]
    # )

    from importlib.util import spec_from_loader, module_from_spec
    from importlib.machinery import SourceFileLoader

    ret = Return(changed=False)

    with CallDepth():
        spec = spec_from_loader(
            playbook, SourceFileLoader(playbook, str(full_playbook_path))
        )
        if spec is None:
            raise ImportError(
                "Unable to spec_from_loader() the module, no error returned."
            )
        module = module_from_spec(spec)
        keys_before = set(module.__dict__.keys())

        up_context.playbook_files_seen.add(str(full_playbook_path))
        spec.loader.exec_module(module)
        up_context.playbook_files_seen.remove(str(full_playbook_path))

        if hoist_vars:
            for key, value in module.__dict__.items():
                if key not in keys_before:
                    module.__dict__[key] = value
                    up_context.playbook_namespace[key] = value

    return ret


@task
def get_url(
    url: TemplateStr, path: TemplateStr, skip_if_path_exists: bool = True
) -> Return:
    """
    Access another playbook from an existing playbook.

    Args:
        url: URL to download.
        path: Location to write file.
        skip_if_path_exists: If `path` already exists, do not download again.

    Examples:

    ```python
    core.get_url(url="https://google.com")
    ```
    """
    if skip_if_path_exists is not None and os.path.exists(path):
        return Return(changed=False)

    r = requests.get(url)
    with open(path, "wb") as fp:
        fp.write(r.content)

    return Return(changed=True)



#!/usr/bin/env python3

"""
Internal uPlaybook objects and methods.

These objects and methods typically are used by task developers, though some may be usful
for reference by users.  These are typically not used directly by end-users, though other
parts of the documentation may refer to these classes.
"""

uplaybook_version = "0.9.5"

import sys
import inspect
from typing import Optional, Union, List, Callable, Any, Iterator
from types import ModuleType
from functools import wraps
import jinja2
import os
import platform
import types
import multiprocessing
import socket
from types import SimpleNamespace
import traceback
import ast
import argparse
import importlib
import pydoc
import re
from pathlib import Path
from collections import namedtuple
import itertools
from rich.console import Console


PlaybookInfo = namedtuple("PlaybookInfo", ["name", "directory", "playbook_file"])


class Failure(Exception):
    """
    The default exception raised if a task fails.
    """

    pass


class Exit(Exception):
    """
    The exception raised to exit a playbook early.
    """

    def __init__(self, msg: str, return_code: int) -> None:
        super().__init__(msg)
        self.return_code = return_code


class PasswordNeeded(Exception):
    """
    An exception which indicates a password is unavailable for decrypting input files.
    """

    pass


def PlatformInfo() -> types.SimpleNamespace:
    """
    See "docs/templating.md#platform_info" for more information.
    """
    env = types.SimpleNamespace()
    uname = platform.uname()
    env.system = platform.system()
    if env.system == "Linux":
        release = platform.freedesktop_os_release()
        env.release_name = release["NAME"]
        env.release_id = release["ID"]
        env.release_version = release["VERSION_ID"]
        env.release_like = release["ID_LIKE"]
        env.release_codename = release["VERSION_CODENAME"]
    if env.system == "Darwin":
        macver = platform.mac_ver()
        env.release_version = macver[0]
    if env.system == "Windows":
        env.release_version = uname.version
        env.release_name = uname.release
        env.release_edition = platform.win32_edition()
    env.arch = uname.machine
    env.cpu_count = multiprocessing.cpu_count()
    env.fqdn = socket.getfqdn()

    try:
        import psutil

        vm = psutil.virtual_memory()
        env.memory_total = vm.total
        env.memory_available = vm.available
        env.memory_used = vm.used
        env.memory_percent_used = vm.percent
    except ImportError:
        pass

    return env


class UpContext:
    """
    A singleton object for storing and managing context data.
    """

    def __init__(self):
        self.globals = {"environ": os.environ, "platform": PlatformInfo()}
        self.context = {"ARGS": SimpleNamespace()}
        self.calling_context = {}
        self.item_context = []
        self.changed_count = 0
        self.failure_count = 0
        self.total_count = 0
        self.ignore_failure_count = 0
        self.handler_list = []
        self.call_depth = 0
        self.remaining_args = []
        self.parsed_args = argparse.Namespace()
        self.playbook_namespace = {}  #  Namespace of the playbook module
        self.playbook_name = ""  #  Name of the playbook being run
        self.playbook_docstring = ""
        self.playbook_directory = "."  #  Directory playbook is in
        self.playbook_files_seen = set()
        self.console = Console()

        self.jinja_env = jinja2.Environment(undefined=jinja2.StrictUndefined)
        self.jinja_env.filters["basename"] = os.path.basename
        self.jinja_env.filters["dirname"] = os.path.dirname
        self.jinja_env.filters["abspath"] = os.path.abspath

    def get_env(self, env_in: Optional[dict] = None) -> dict:
        """Returns the jinja template environment"""
        env = self.globals.copy()
        env.update(self.context)
        env.update(self.playbook_namespace)
        env.update(self.calling_context)
        for ctx in self.item_context[::-1]:
            env.update(ctx)
        if env_in:
            env.update(env_in)
        return env

    def ignore_failures(self):
        """Is ignore_failures mode active?"""
        return self.ignore_failure_count > 0

    def add_handler(self, fn: Callable) -> None:
        """Add a notify function."""
        if fn in self.handler_list:
            return

        self.handler_list.append(fn)

    def flush_handlers(self) -> None:
        """Run all the handler functions."""
        did_handler = False
        while self.handler_list:
            did_handler = True
            fn = self.handler_list.pop(0)
            print(f">> *** Starting handler: {fn.__name__}")
            fn()
        if did_handler:
            print(f">> *** Done with handlers")


up_context = UpContext()


class CallDepth:
    """
    A context manager to increment the call depth when one task calls another task.
    """

    def __enter__(self):
        up_context.call_depth += 1

    def __exit__(self, *args):
        up_context.call_depth -= 1


class TemplateStr(str):
    """
    A subclass of str to mark what arguments of a task are templated.
    """

    pass


class RawStr(str):
    """
    A subclass of str which is not to have Jinja2 template expansion.
    """

    pass


def template_args(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    Decorator to pre-processes arguments to the function of type TemplateStr through Jinja.

    If an argument has a type annotation of TemplateStr this decorator will treat the string
    value of that argument as a Jinja2 template and render it. The resulting value (after
    rendering) is then passed to the wrapped function.

    Args:
        func (Callable): The function to be wrapped.

    Returns:
        Callable: The wrapped function with potentially modified arguments.

    Usage:
    @template_args
    def example_function(a: Union[TemplateStr, str]):
        print(a)

    example_function("Hello {{ 'World' }}")  # Outputs: "Hello World"
    """
    sig = inspect.signature(func)

    def _render_jinja_arg(s: str) -> str:
        """Render the arguments as Jinja2, use the up_context and the calling environment.
        NOTE: This is hardcoded to be run from inside this decorator
        Is likely to be fragile.
        """
        if type(s) == RawStr:
            return s
        return up_context.jinja_env.from_string(s).render(up_context.get_env())

    @wraps(func)
    def wrapper(*args, **kwargs):
        # Convert args to mutable list
        args = list(args)

        # Get bound arguments
        bound_args = sig.bind(*args, **kwargs)
        bound_args.apply_defaults()

        # Process bound arguments and replace if type is TemplateStr
        for name, value in bound_args.arguments.items():
            if value not in args and name not in kwargs:
                continue
            if not (
                isinstance(value, list) and isinstance(value[0], str)
            ) and not isinstance(value, str):
                continue

            annotation = sig.parameters[name].annotation

            if (
                annotation == Optional[List[TemplateStr]]
                or annotation == List[TemplateStr]
            ) and isinstance(value, list):
                if name in kwargs:
                    kwargs[name] = list([_render_jinja_arg(x) for x in value])
                else:
                    args[bound_args.args.index(value)] = list(
                        [_render_jinja_arg(x) for x in value]
                    )

            # Check for TemplateStr directly or as part of a Union
            elif isinstance(value, str) and (
                annotation is TemplateStr
                or (
                    hasattr(annotation, "__origin__")
                    and issubclass(TemplateStr, annotation.__args__)
                )
            ):
                if name in kwargs:
                    kwargs[name] = _render_jinja_arg(value)
                else:
                    args[bound_args.args.index(value)] = _render_jinja_arg(value)

        return func(*args, **kwargs)

    return wrapper


TaskCallInfo = namedtuple(
    "TaskCallInfo",
    [
        "function_name",
        "function_qualname",
        "module_name",
        "annotations",
        "args",
        "kwargs",
    ],
)


def calling_context(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    Decorator to save off the calling function namespace into up_context.calling_context.

    This decorator saves off the namespace of the function that calls the wrapped function
    for later use by templating.

    Args:
        func (Callable): The function to be wrapped.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        current_frame = inspect.currentframe()
        env = current_frame.f_back.f_locals.copy()
        up_context.calling_context = env
        up_context.task_call_info = TaskCallInfo(
            func.__name__,
            func.__qualname__,
            func.__module__,
            func.__annotations__,
            args,
            kwargs.copy(),
        )

        ret = func(*args, **kwargs)

        up_context.task_call_info = None
        up_context.calling_context = {}

        return ret

    return wrapper


def task(func: Callable[..., Any]) -> Callable[..., Any]:
    """A decorator for tasks, combines @calling_context and @template_args"""
    func.__is_uplaybook_task__ = True
    return calling_context(template_args(func))


class Return:
    """
    A return type from tasks to track success/failure, display status, etc...

    This can be used as a context manager if a `context_manager` function is provided.
    See `fs.cd()` for an example.

    It can also be used as a boolean, to check success/failure (needs `ignore_failure`, or failure
    will raise exception).  See `core.run()` for an exaple.

    Args:
        changed: If True, mark the task as having changed system state. (bool)
        failure: If True, mark the task as having failed. (optional, bool)
        ignore_failure: If True, failure is not considered fatal, execution can continue on.
                (bool, default False)
        extra_message: An extra message to display in the status line in parens (optional, str)
        output: Output of the task to display, for example stdout of a run command.  (optional, str)
        hide_args: If true, do not display any arguments names/values.  Useful for `debug()`
                which doesn't need to show the message in the status line as well as the output.
                (bool, default=False)
        secret_args: Arguments that have secrets in them, so obscure the value.  (optional, set)
        extra: Extra data to be returned to the caller, as a `types.SimpleNamespace()`, for
                example `run()` will return exit code and stderr as extra.  (optional, types.SimpleNamespace())
        raise_exc: Raise this exception after handling the return.  If failure=True, this exception will be
                raised, if not specified a Failure() exception will be raised. (optional, Exception())
        context_manager: This type can optionally behave as a Context Manager, and if so this function
                will be called with no parameters at the end of the context.  Use a closure if you want to
                associate data with the function call ("lambda: function(args)").  (optional, Callable).

    Examples:

        Return(changed=True)
        Return(changed=False)
        Return(changed=True, extra_message="Permissions")
        Return(changed=True, extra=SimpleNamespace(stderr=s))
        Return(changed=True, context_manager=lambda: f(arg))
    """

    def __init__(
        self,
        changed: bool,
        failure: bool = False,
        ignore_failure: bool = False,
        extra_message: Optional[str] = None,
        output: Optional[str] = None,
        hide_args: bool = False,
        secret_args: set = set(),
        extra: Optional[SimpleNamespace] = None,
        raise_exc: Optional[Exception] = None,
        context_manager: Optional[Callable] = None,
    ) -> None:
        self.changed = changed
        self.extra_message = extra_message
        self.output = output
        self.hide_args = hide_args
        self.extra = extra
        self.failure = failure
        self.secret_args = secret_args
        self.raise_exc = raise_exc
        self.context_manager = context_manager

        self.print_status()

        failure_ok = ignore_failure or up_context.ignore_failures is True

        up_context.total_count += 1
        if changed:
            up_context.changed_count += 1
        if failure and not failure_ok:
            up_context.failure_count += 1
            raise self.raise_exc if self.raise_exc is not None else Failure(
                "Unspecified failure in task"
            )
        if self.raise_exc:
            raise self.raise_exc

    def __enter__(self) -> "Return":
        """
        Begin a context if used as a context manager.

        Raises:
            AttributeError: if no `context_manager` was specified.

        Returns:
            The Return() object, so it can be used as "with fn() as return_obj:"
        """
        if not self.context_manager:
            raise AttributeError("This instance is not a valid context manager.")
        return self

    def __exit__(self, *_) -> None:
        """
        End the context if used as a context manager.

        This will call the `context_manager` function if it is given.
        """
        assert self.context_manager is not None
        self.context_manager()

    def print_status(self) -> None:
        """
        Display the output and status of the task.
        """
        parent_function_name = "<Unknown>"
        for parent_frame_info in inspect.stack()[2:]:
            parent_function_name = parent_frame_info.function
            if not parent_function_name.startswith("_"):
                break

        if self.hide_args:
            call_args = "..."
        else:
            args, _, _, values = inspect.getargvalues(parent_frame_info.frame)

            #  overwrite the original arguments (if any had been modified in function call)
            #  NOTE: This only works for the inner-most of nested calls, this will need to
            #  be converted to a stack to handle nesting.
            if up_context.task_call_info:
                values.update(up_context.task_call_info.kwargs)

            call_args = ", ".join(
                [
                    f"{arg}=" + ("***" if arg in self.secret_args else f"{values[arg]}")
                    for arg in args
                    if arg != "self" and values[arg] is not None
                ]
            )

        add_msg = f" ({self.extra_message})" if self.extra_message else ""

        prefix = "=#"
        suffix = ""
        style = ""
        if self.changed:
            prefix = "=>"
            style = "green"
        if self.failure:
            prefix = "=!"
            suffix = " (failure ignored)"
            style = "red"
        call_depth = "=" * up_context.call_depth

        up_context.console.print(
            f"{call_depth}{prefix} {parent_function_name}({call_args}){add_msg}{suffix}",
            style=style,
            highlight=False,
            markup=False,
        )
        if self.output:
            up_context.console.print(self.output, markup=False, highlight=True)

    def __repr__(self) -> str:
        """
        Format this object for display.
        """
        values = [f"changed={self.changed}"]
        if self.extra_message is not None:
            values.append(f"extra_message={repr(self.extra_message)}")
        if self.extra is not None:
            for k, v in vars(self.extra).items():
                values.append(f"extra.{k}={repr(v)}")

        formatted_output = ""
        if self.output is not None:
            if len(self.output) < 60 or self.output.count("\n") < 4:
                values.append(f"output={repr(self.output)}")
            else:
                formatted_output = ',\noutput="""\n' + self.output + '"""'
        return f"Return({', '.join(values)}{formatted_output})"

    def __bool__(self) -> bool:
        """
        If checked for truthfulness, return True if not `failure`.
        """
        return not self.failure

    def notify(self, handler: Union[Callable, List]) -> None:
        """
        Register a handler function to call if changed.

        Args:
            handler:  A function or a list of functions to register for calling later,
                    if the task has changed the system.  (callable or list)
        """
        if self.changed:
            if callable(handler):
                up_context.add_handler(handler)
            else:
                for x in handler:
                    up_context.add_handler(x)


def extract_docstring_from_file(filename: str) -> Union[str, None]:
    """Open the specified file and retrieve the docstring from it.

    Returns None if no docstring is found"""
    if os.path.isdir(filename):
        filename = os.path.join(filename, "playbook")
    with open(filename, "r") as f:
        try:
            node = ast.parse(f.read())
        except Exception:
            return None

        if (
            node.body
            and isinstance(node.body[0], ast.Expr)
            and isinstance(node.body[0].value, ast.Str)
        ):
            return ast.get_docstring(node)

    return None


def import_script_as_module(module_name: str, paths_to_try: List[str]) -> ModuleType:
    """
    Imports a Python script as a module, whether it ends in ".py" or not.
    Given the name of a module to import, and a list of absolute or relative path names
    (including the filename), import the module.  The module is set up so that it can
    later be imported, but a reference to the module is also returned.

    Args:
        module_name (str): The name of the module to import.
                This doesn't have to match the filename.
        paths_to_try (List[str]): A list of file paths to look for the file to load.
                This can be absolute or relative paths to the file, the first file that
                exists is used.

    Returns:
        Module: A reference to the imported module.

    Raises:
        FileNotFoundError: If the module file is not found in any of the specified directory paths.
        ImportError: If there are issues importing the module, such as invalid Python syntax in the module file.

    Example:
        my_module = import_script_as_module("my_module", ["my_module", "../my_module"])

        # Now you can either directly use "my_module"
        my_module.function()

        # Or you can later import it:
        import my_module
    """
    from pathlib import Path
    import os

    for try_filename in paths_to_try:
        if os.path.exists(try_filename):
            module_filename = Path(try_filename).resolve()
            break
    else:
        raise FileNotFoundError(f"Unable to find '{module_name}' module to import")

    from importlib.util import spec_from_loader, module_from_spec
    from importlib.machinery import SourceFileLoader
    import sys

    spec = spec_from_loader(
        module_name, SourceFileLoader(module_name, str(module_filename))
    )
    if spec is None:
        raise ImportError("Unable to spec_from_loader() the module, no error returned.")
    module = module_from_spec(spec)
    module.__dict__.update(up_context.get_env())
    up_context.playbook_namespace = module.__dict__
    spec.loader.exec_module(module)

    return module


def find_updocs(name: str) -> str:
    """
    Given a module/task name, return the docstring for it.

    This iterates over the various options for how a "--up-docs" may be requested and returns
    the first appropriate docstring found.

    Args:
        name: The "--up-docs" document to find.

    Returns:
        The associated updoc.
    """
    if name == "__main__":
        from . import __doc__

        return __doc__

    #  if a module is specified
    try:
        module = importlib.import_module(f".{name}", package=__package__)
        docs = (module.__doc__ if module.__doc__ is not None else "").rstrip()

        task_functions = []
        for attr_name in dir(module):
            attr = getattr(module, attr_name)
            if (
                callable(attr)
                and getattr(attr, "__doc__", None)
                and getattr(attr, "__is_uplaybook_task__", False)
            ):
                first_line = getattr(attr, "__doc__", "").lstrip().split("\n")[0]
                task_functions.append(f"{attr_name} - {first_line}")

        docs = (docs + "\n\n## Available Tasks:\n\n").lstrip("\n")
        for task_name in task_functions:
            docs += f"- {name}.{task_name}\n"

        return docs
    except ModuleNotFoundError:
        pass

    #  module.task docs
    module_name, function_name = name.rsplit(".", 1)
    module = importlib.import_module(f".{module_name}", package=__package__)
    function = getattr(module, function_name)
    return f"# {name}\n\n" + function.__doc__.lstrip("\n")


def display_docs(name: str) -> None:
    """
    Display the documentation for the component specified by `name`.

    This displays the docstring from the module or task referred to by `name`,
    formatting it for nicer representation.

    Args:
        name: Module name or module.task name.  If special value "__main__" it displays
                the up2 documentation.
    """
    docs = find_updocs(name)
    pydoc.pager(re.sub(r"#\w+", "", docs).rstrip())


class UpArgumentParser(argparse.ArgumentParser):
    """Wrapper so that "up", when run with no arguments, lists available playbooks."""

    def print_usage(self, file=None):
        super().print_usage(file)
        self.help_list_playbooks()

    def print_help(self, file=None):
        super().print_help(file)
        self.help_list_playbooks()

    def help_list_playbooks(self):
        print()
        print("Available playbooks:")
        playbooks_seen = set()
        for playbook in list_playbooks():
            duplicate = (
                " *HIDDEN BY PREVIOUS PLAYBOOK*"
                if playbook.name in playbooks_seen
                else ""
            )
            playbooks_seen.add(playbook.name)
            print(f"  - {playbook.name} ({playbook.directory}{duplicate})")
            docs = extract_docstring_from_file(playbook.playbook_file)
            desc = (docs if docs else "").lstrip().split("\n", 1)[0]

            if desc:
                print(f"      {desc}")
        print()


def parse_args() -> argparse.Namespace:
    """
    The main CLI argument parser.

    Returns:
        The parsed arguments.
    """
    parser = UpArgumentParser(
        prog="up",
        description="Run playbooks of actions, typically to set up some sort of environment.",
        add_help=False,
    )

    parser.add_argument(
        "--help",
        action="store_true",
        help="Display help about usage.",
    )
    parser.add_argument(
        "--up-full-traceback",
        action="store_true",
        help="Display a full traceback rather than the default abbreviated version.",
    )
    parser.add_argument(
        "--up-list-playbooks",
        action="store_true",
        help="Display a list of playbook names and exit.",
    )
    parser.add_argument(
        "--up-debug",
        action="store_true",
        help="Display additional debugging information during playbook run.",
    )
    parser.add_argument(
        "--up-version",
        action="store_true",
        help="Display the version and exit?",
    )
    parser.add_argument(
        "--up-docs",
        type=str,
        nargs="?",
        const="__main__",
        default=None,
        dest="docs_arg",
        help="Display documentation, if an optional value is given the help for that component will be displayed.",
    )
    parser.add_argument(
        "playbook", type=str, nargs="?", default=None, help="Name of the playbook."
    )

    if "--up-version" in sys.argv:
        print(f"uPlaybook version: {uplaybook_version}")
        sys.exit(0)

    args, remaining_args = parser.parse_known_args()

    if args.up_list_playbooks:
        for playbook in list_playbooks():
            print(f"{playbook.name}")
        sys.exit(0)
    if args.help:
        if args.playbook is None:
            parser.print_help()
            sys.exit(0)
        remaining_args.insert(0, "--help")
    if args.docs_arg:
        display_docs(args.docs_arg)
        if args.docs_arg == "__main__":
            print()
            parser.print_help()
        sys.exit(0)

    if not args.playbook:
        parser.print_usage()
        sys.exit(1)

    up_context.remaining_args = remaining_args
    up_context.parsed_args = args

    return args


def get_playbook_search_paths() -> List[Path]:
    """
    Get the playbook search path (either the default or from the environment)
    and return an iterable of the path objects.

    Returns:
        List[Paths] for the list of locations to look for playbooks."""

    search_path = os.environ.get(
        "UP_PLAYBOOK_PATH",
        ".:.uplaybooks:~/.config/uplaybook:~/.config/uplaybook/library:/etc/uplaybook",
    )
    return [Path(x).expanduser().joinpath(".") for x in search_path.split(":")]


def list_playbooks() -> Iterator[PlaybookInfo]:
    """
    Walk the playbook path and return a list of available playbooks.
    Playbook files take precedence over "playbook/playbook".  Sorted by
    playbook name within each component of the search path.

    Returns:

    """
    for playbook_path in get_playbook_search_paths():
        possible_playbooks = sorted(
            itertools.chain(
                playbook_path.glob("*.pb"), playbook_path.glob("*/playbook")
            ),
            key=lambda x: x.name,
        )
        for playbook_file in possible_playbooks:
            if playbook_file.exists():
                directory = playbook_file.parent
                if playbook_file.name == "playbook" and directory.as_posix() != ".":
                    name = directory.name
                else:
                    name = playbook_file.name
                yield PlaybookInfo(name, directory, playbook_file)


def find_playbook(playbookname: str) -> PlaybookInfo:
    """
    Finds and returns the path of a specified playbook file.

    Search for the playbook in the UP_PLAYBOOK_PATH environment variable,
    or a default if not specified.

    Args:
        playbookname (str): The name of the playbook file to search for.

    Returns:
        Path: The path of the found playbook file.

    Raises:
        FileNotFoundError: If the playbook file is not found in the search paths."""

    #  absolute path name given, just use it
    if os.path.sep in playbookname or (
        os.path.altsep and os.path.altsep in playbookname
    ):
        playbook = Path(playbookname)
        if playbook.is_dir() and (playbook / "playbook").exists():
            playbook = playbook / "playbook"
        return PlaybookInfo(
            playbook.name, playbook.absolute().parent, playbook.absolute()
        )

    for playbook in list_playbooks():
        if playbook.name == playbookname or (
            playbook.name.endswith(".pb")
            and not playbookname.endswith(".pb")
            and playbook.name == (playbookname + ".pb")
        ):
            return playbook

    searchpath = get_playbook_search_paths()
    raise FileNotFoundError(
        f"Unable to locate a playbook by the name of {playbookname},"
        f" searched in path {searchpath}."
    )


def find_file(filename: str) -> Path:
    """
    Finds and returns the path of a template/file.

    This function uses a colon-separated search path, either gotten from the
    UP_FILES_PATH environment variable or the default.  "..." specified in
    the search path is relative to the directory the playbook is found in.

    Returns:
    Path: The path of the found template file.

    Raises:
    FileNotFoundError: If the template file is not found in the search paths.
    """
    search_path = os.environ.get("UP_FILES_PATH", "...:.../files:.")

    if Path(filename).is_absolute():
        return Path(filename)

    playbook_directory = Path(up_context.playbook_directory)

    for directory in search_path.split(":"):
        if directory == "...":
            p = playbook_directory.joinpath(filename)
            if p.exists():
                return p
            continue
        if directory.startswith(".../"):
            p = playbook_directory.joinpath(directory[4:]).joinpath(filename)
            if p.exists():
                return p
            continue

        p = Path(directory).joinpath(filename)
        if p.exists():
            return p

    raise FileNotFoundError(
        f"Could not find file {filename}, searched in {search_path}"
    )


def show_playbook_traceback(e: Exception) -> None:
    """
    Display the current traceback, but only the parts that reference the playbook(s).
    Traceback is printed to stdout.
    """
    up_context.console.print(
        "[bold red]uPlaybook Traceback (most recent call last):[/]"
    )
    exc_type, exc_value, exc_tb = sys.exc_info()
    for entry in traceback.extract_tb(exc_tb):
        if entry.filename not in up_context.playbook_files_seen:
            continue
        up_context.console.print(
            f'  File: "{entry.filename}", line {entry.lineno}, in {entry.name}',
            highlight=True,
        )
        up_context.console.print(f"    {entry.line}", highlight=True)
        #', '_line', 'filename', 'line', 'lineno', 'locals', 'name']

    if isinstance(e, Failure):
        up_context.console.print(
            f"[bold red]Task Failure, Cause:[/]\n   {str(e)}", highlight=True
        )
    else:
        up_context.console.print(
            "".join(traceback.format_exception_only(exc_value)).rstrip(), highlight=True
        )


def cli() -> None:
    """
    The main entry point for the CLI.
    """
    args = parse_args()

    full_playbook_path = None
    return_code = 0
    try:
        pb_name = args.playbook
        up_context.playbook_name = pb_name
        playbook = find_playbook(pb_name)
        up_context.playbook_directory = playbook.directory.absolute()
        full_playbook_path = Path(playbook.playbook_file).absolute()

        docs = extract_docstring_from_file(full_playbook_path)
        up_context.playbook_docstring = docs if docs else ""

        up_context.playbook_files_seen.add(full_playbook_path.as_posix())
        import_script_as_module(
            pb_name, [playbook.playbook_file, playbook.playbook_file]
        )
    except Exit as e:
        return_code = e.return_code
        pass
    except Exception as e:
        if args.up_full_traceback or not full_playbook_path:
            print(traceback.format_exc())
        else:
            show_playbook_traceback(e)
        return_code = 1

    up_context.flush_handlers()

    recap_msg = "*** RECAP"
    failure_count_msg = f"failure={up_context.failure_count}"
    if up_context.failure_count > 0:
        failure_count_msg = f"[bold red]{failure_count_msg}[/]"
        recap_msg = f"[bold red]{recap_msg}[/]"
    else:
        recap_msg = f"[green]{recap_msg}[/]"

    up_context.console.print(
        f"\n{recap_msg}:  total={up_context.total_count} changed={up_context.changed_count} {failure_count_msg}",
        markup=True,
        highlight=True,
    )

    sys.exit(return_code)



#!/usr/bin/env python3

"""
## brew


Manage brew packages on mac/OSX. See https://brew.sh/
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def update():
    """
    Updates brew repositories.
    """
    operargs = {}

    result = _run_pyinfra(
        "from pyinfra.operations import brew", "brew.update", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def upgrade():
    """
    Upgrades all brew packages.
    """
    operargs = {}

    result = _run_pyinfra(
        "from pyinfra.operations import brew", "brew.upgrade", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def packages(packages=None, present=True, latest=False, update=False, upgrade=False):
    """
    Add/remove/update brew packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + update: run ``brew update`` before installing packages
    + upgrade: run ``brew upgrade`` before installing packages

    Versions:
        Package versions can be pinned like brew: ``<pkg>@<version>``.

    **Examples:**

    .. code:: python

        # Update package list and install packages
        brew.packages(
            name='Install Vim and vimpager',
            packages=["vimpager", "vim"],
            update=True,
        )

        # Install the latest versions of packages (always check)
        brew.packages(
            name="Install latest Vim",
            packages=["vim"],
            latest=True,
        )
    """
    operargs = {
        "packages": repr(packages),
        "present": repr(present),
        "latest": repr(latest),
        "update": repr(update),
        "upgrade": repr(upgrade),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import brew", "brew.packages", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def cask_args(host):
    operargs = {
        "host": repr(host),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import brew", "brew.cask_args", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def cask_upgrade():
    """
    Upgrades all brew casks.
    """
    operargs = {}

    result = _run_pyinfra(
        "from pyinfra.operations import brew", "brew.cask_upgrade", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def casks(casks=None, present=True, latest=False, upgrade=False):
    """
    Add/remove/update brew casks.

    + casks: list of casks to ensure
    + present: whether the casks should be installed
    + latest: whether to upgrade casks without a specified version
    + upgrade: run brew cask upgrade before installing casks

    Versions:
        Cask versions can be pinned like brew: ``<pkg>@<version>``.

    **Example:**

    .. code:: python

        brew.casks(
            name='Upgrade and install the latest cask',
            casks=["godot"],
            upgrade=True,
            latest=True,
        )
    """
    operargs = {
        "casks": repr(casks),
        "present": repr(present),
        "latest": repr(latest),
        "upgrade": repr(upgrade),
    }

    result = _run_pyinfra("from pyinfra.operations import brew", "brew.casks", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def tap(src, present=True):
    """
    Add/remove brew taps.

    + src: the name of the tap
    + present: whether this tap should be present or not

    **Examples:**

    .. code:: python

        brew.tap(
            name="Add a brew tap",
            src="includeos/includeos",
        )

        # Multiple taps
        for tap in ["includeos/includeos", "ktr0731/evans"]:
            brew.tap(
                name={f"Add brew tap {tap}"},
                src=tap,
            )
    """
    operargs = {
        "src": repr(src),
        "present": repr(present),
    }

    result = _run_pyinfra("from pyinfra.operations import brew", "brew.tap", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## Systemd tasks

This module provides tasks for interacting with systemd.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def daemon_reload(user_mode=False, machine=None, user_name=None):
    """
    Reload the systemd daemon to read unit file changes.

    + user_mode: whether to use per-user systemd (systemctl --user) or not
    + machine: the machine name to connect to
    + user_name: connect to a specific user's systemd session
    """
    operargs = {
        "user_mode": repr(user_mode),
        "machine": repr(machine),
        "user_name": repr(user_name),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import systemd", "systemd.daemon_reload", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def service(
    service,
    running=True,
    restarted=False,
    reloaded=False,
    command=None,
    enabled=None,
    daemon_reload=False,
    user_mode=False,
    machine=None,
    user_name=None,
):
    """
    Manage the state of systemd managed units.

    + service: name of the systemd unit to manage
    + running: whether the unit should be running
    + restarted: whether the unit should be restarted
    + reloaded: whether the unit should be reloaded
    + command: custom command to pass like: ``/etc/rc.d/<name> <command>``
    + enabled: whether this unit should be enabled/disabled on boot
    + daemon_reload: reload the systemd daemon to read updated unit files
    + user_mode: whether to use per-user systemd (systemctl --user) or not
    + machine: the machine name to connect to
    + user_name: connect to a specific user's systemd session

    **Examples:**

    .. code:: python

        systemd.service(
            name="Restart and enable the dnsmasq service",
            service="dnsmasq.service",
            running=True,
            restarted=True,
            enabled=True,
        )

        systemd.service(
            name="Enable logrotate timer",
            service="logrotate.timer",
            running=True,
            enabled=True,
        )
    """
    operargs = {
        "service": repr(service),
        "running": repr(running),
        "restarted": repr(restarted),
        "reloaded": repr(reloaded),
        "command": repr(command),
        "enabled": repr(enabled),
        "daemon_reload": repr(daemon_reload),
        "user_mode": repr(user_mode),
        "machine": repr(machine),
        "user_name": repr(user_name),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import systemd", "systemd.service", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## sysvinit


Manage sysvinit services (``/etc/init.d``).
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def service(
    service, running=True, restarted=False, reloaded=False, enabled=None, command=None
):
    """
    Manage the state of SysV Init (/etc/init.d) services.

    + service: name of the service to manage
    + running: whether the service should be running
    + restarted: whether the service should be restarted
    + reloaded: whether the service should be reloaded
    + enabled: whether this service should be enabled/disabled
    + command: command (eg. reload) to run like: ``/etc/init.d/<service> <command>``

    Enabled:
        Because managing /etc/rc.d/X files is a mess, only certain Linux distributions
        support enabling/disabling services:

        + Ubuntu/Debian (``update-rc.d``)
        + CentOS/Fedora/RHEL (``chkconfig``)
        + Gentoo (``rc-update``)

        For other distributions and more granular service control, see the
        ``sysvinit.enable`` operation.

    **Example:**

    .. code:: python

        sysvinit.service(
            name="Restart and enable rsyslog",
            service="rsyslog",
            restarted=True,
            enabled=True,
        )
    """
    operargs = {
        "service": repr(service),
        "running": repr(running),
        "restarted": repr(restarted),
        "reloaded": repr(reloaded),
        "enabled": repr(enabled),
        "command": repr(command),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import sysvinit", "sysvinit.service", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def enable(
    service,
    start_priority=20,
    stop_priority=80,
    start_levels=(2, 3, 4, 5),
    stop_levels=(0, 1, 6),
):
    """
    Manually enable /etc/init.d scripts by creating /etc/rcX.d/Y links.

    + service: name of the service to enable
    + start_priority: priority to start the service
    + stop_priority: priority to stop the service
    + start_levels: which runlevels should the service run when enabled
    + stop_levels: which runlevels should the service stop when enabled

    **Example:**

    .. code:: python

        init.d_enable(
            name="Finer control on which runlevels rsyslog should run",
            service="rsyslog",
            start_levels=(3, 4, 5),
            stop_levels=(0, 1, 2, 6),
        )
    """
    operargs = {
        "service": repr(service),
        "start_priority": repr(start_priority),
        "stop_priority": repr(stop_priority),
        "start_levels": repr(start_levels),
        "stop_levels": repr(stop_levels),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import sysvinit", "sysvinit.enable", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## Server tasks

This module provides tasks for working with OS services.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def reboot(delay=10, interval=1, reboot_timeout=300):
    """
    Reboot the server and wait for reconnection.

    + delay: number of seconds to wait before attempting reconnect
    + interval: interval (s) between reconnect attempts
    + reboot_timeout: total time before giving up reconnecting

    **Example:**

    .. code:: python

        server.reboot(
            name="Reboot the server and wait to reconnect",
            delay=60,
            reboot_timeout=600,
        )
    """
    operargs = {
        "delay": repr(delay),
        "interval": repr(interval),
        "reboot_timeout": repr(reboot_timeout),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.reboot", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def wait(port):
    """
    Waits for a port to come active on the target machine. Requires netstat, checks every
    second.

    + port: port number to wait for

    **Example:**

    .. code:: python

        server.wait(
            name="Wait for webserver to start",
            port=80,
        )
    """
    operargs = {
        "port": repr(port),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.wait", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def shell(commands):
    """
    Run raw shell code on server during a deploy. If the command would
    modify data that would be in a fact, the fact would not be updated
    since facts are only run at the start of a deploy.

    + commands: command or list of commands to execute on the remote server

    **Example:**

    .. code:: python

        server.shell(
            name="Run lxd auto init",
            commands=["lxd init --auto"],
        )
    """
    operargs = {
        "commands": repr(commands),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.shell", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def script(src, args=()):
    """
    Upload and execute a local script on the remote host.

    + src: local script filename to upload & execute
    + args: iterable to pass as arguments to the script

    **Example:**

    .. code:: python

        # Note: This assumes there is a file in files/hello.bash locally.
        server.script(
            name="Hello",
            src="files/hello.bash",
        )

        # Example passing arguments to the script
        server.script(
            name="Hello",
            src="files/hello.bash",
            args=("do-something", "with-this"),
        )
    """
    operargs = {
        "src": repr(src),
        "args": repr(args),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.script", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def script_template(src, args=(), **data):
    """
    Generate, upload and execute a local script template on the remote host.

    + src: local script template filename

    **Example:**

    .. code:: python

        # Example showing how to pass python variable to a script template file.
        # The .j2 file can use `{{ some_var }}` to be interpolated.
        # To see output need to run pyinfra with '-v'
        # Note: This assumes there is a file in templates/hello2.bash.j2 locally.
        some_var = 'blah blah blah '
        server.script_template(
            name="Hello from script",
            src="templates/hello2.bash.j2",
            some_var=some_var,
        )
    """
    operargs = {
        "src": repr(src),
        "args": repr(args),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.script_template", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def modprobe(module, present=True, force=False):
    """
    Load/unload kernel modules.

    + module: name of the module to manage
    + present: whether the module should be loaded or not
    + force: whether to force any add/remove modules

    **Example:**

    .. code:: python

        server.modprobe(
            name="Silly example for modprobe",
            module="floppy",
        )
    """
    operargs = {
        "module": repr(module),
        "present": repr(present),
        "force": repr(force),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.modprobe", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def mount(path, mounted=True, options=None, device=None, fs_type=None):
    """
    Manage mounted filesystems.

    + path: the path of the mounted filesystem
    + mounted: whether the filesystem should be mounted
    + options: the mount options

    Options:
        If the currently mounted filesystem does not have all of the provided
        options it will be remounted with the options provided.

    ``/etc/fstab``:
        This operation does not attempt to modify the on disk fstab file - for
        that you should use the `files.line operation <./files.html#files-line>`_.
    """
    operargs = {
        "path": repr(path),
        "mounted": repr(mounted),
        "options": repr(options),
        "device": repr(device),
        "fs_type": repr(fs_type),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.mount", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def hostname(hostname, hostname_file=None):
    """
    Set the system hostname using ``hostnamectl`` or ``hostname`` on older systems.

    + hostname: the hostname that should be set
    + hostname_file: the file that permanently sets the hostname

    Hostname file:
        The hostname file only matters no systems that do not have ``hostnamectl``,
        which is part of ``systemd``.

        By default pyinfra will auto detect this by targeting ``/etc/hostname``
        on Linux and ``/etc/myname`` on OpenBSD.

        To completely disable writing the hostname file, set ``hostname_file=False``.

    **Example:**

    .. code:: python

        server.hostname(
            name="Set the hostname",
            hostname="server1.example.com",
        )
    """
    operargs = {
        "hostname": repr(hostname),
        "hostname_file": repr(hostname_file),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.hostname", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def sysctl(key, value, persist=False, persist_file="/etc/sysctl.conf"):
    """
    Edit sysctl configuration.

    + key: name of the sysctl setting to ensure
    + value: the value or list of values the sysctl should be
    + persist: whether to write this sysctl to the config
    + persist_file: file to write the sysctl to persist on reboot

    **Example:**

    .. code:: python

        server.sysctl(
            name="Change the fs.file-max value",
            key="fs.file-max",
            value=100000,
            persist=True,
        )
    """
    operargs = {
        "key": repr(key),
        "value": repr(value),
        "persist": repr(persist),
        "persist_file": repr(persist_file),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.sysctl", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def service(
    service, running=True, restarted=False, reloaded=False, command=None, enabled=None
):
    """
    Manage the state of services. This command checks for the presence of all the
    Linux init systems ``pyinfra`` can handle and executes the relevant operation.

    + service: name of the service to manage
    + running: whether the service should be running
    + restarted: whether the service should be restarted
    + reloaded: whether the service should be reloaded
    + command: custom command execute
    + enabled: whether this service should be enabled/disabled on boot

    **Example:**

    .. code:: python

        server.service(
            name="Enable open-vm-tools service",
            service="open-vm-tools",
            enabled=True,
        )
    """
    operargs = {
        "service": repr(service),
        "running": repr(running),
        "restarted": repr(restarted),
        "reloaded": repr(reloaded),
        "command": repr(command),
        "enabled": repr(enabled),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.service", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def packages(packages, present=True):
    """
    Add or remove system packages. This command checks for the presence of all the
    system package managers ``pyinfra`` can handle and executes the relevant operation.

    + packages: list of packages to ensure
    + present: whether the packages should be installed

    **Example:**

    .. code:: python

        server.packages(
            name="Install Vim and vimpager",
            packages=["vimpager", "vim"],
        )
    """
    operargs = {
        "packages": repr(packages),
        "present": repr(present),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.packages", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def crontab(
    command,
    present=True,
    user=None,
    cron_name=None,
    minute="*",
    hour="*",
    month="*",
    day_of_week="*",
    day_of_month="*",
    special_time=None,
    interpolate_variables=False,
):
    """
    Add/remove/update crontab entries.

    + command: the command for the cron
    + present: whether this cron command should exist
    + user: the user whose crontab to manage
    + cron_name: name the cronjob so future changes to the command will overwrite
    + minute: which minutes to execute the cron
    + hour: which hours to execute the cron
    + month: which months to execute the cron
    + day_of_week: which day of the week to execute the cron
    + day_of_month: which day of the month to execute the cron
    + special_time: cron "nickname" time (@reboot, @daily, etc), overrides others
    + interpolate_variables: whether to interpolate variables in ``command``

    Cron commands:
        Unless ``name`` is specified the command is used to identify crontab entries.
        This means commands must be unique within a given users crontab. If you require
        multiple identical commands, provide a different name argument for each.

    Special times:
        When provided, ``special_time`` will be used instead of any values passed in
        for ``minute``/``hour``/``month``/``day_of_week``/``day_of_month``.

    **Example:**

    .. code:: python

        # simple example for a crontab
        server.crontab(
            name="Backup /etc weekly",
            command="/bin/tar cf /tmp/etc_bup.tar /etc",
            name="backup_etc",
            day_of_week=0,
            hour=1,
            minute=0,
        )
    """
    operargs = {
        "command": repr(command),
        "present": repr(present),
        "user": repr(user),
        "cron_name": repr(cron_name),
        "minute": repr(minute),
        "hour": repr(hour),
        "month": repr(month),
        "day_of_week": repr(day_of_week),
        "day_of_month": repr(day_of_month),
        "special_time": repr(special_time),
        "interpolate_variables": repr(interpolate_variables),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.crontab", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def group(group, present=True, system=False, gid=None):
    """
    Add/remove system groups.

    + group: name of the group to ensure
    + present: whether the group should be present or not
    + system: whether to create a system group
    + gid: use a specific groupid number

    System users:
        System users don't exist on BSD, so the argument is ignored for BSD targets.

    **Examples:**

    .. code:: python

        server.group(
            name="Create docker group",
            group="docker",
        )

        # multiple groups
        for group in ["wheel", "lusers"]:
            server.group(
                name=f"Create the group {group}",
                group=group,
            )
    """
    operargs = {
        "group": repr(group),
        "present": repr(present),
        "system": repr(system),
        "gid": repr(gid),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.group", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def user_authorized_keys(
    user,
    public_keys,
    group=None,
    delete_keys=False,
    authorized_key_directory=None,
    authorized_key_filename=None,
):
    """
    Manage `authorized_keys` of system users.

    + user: name of the user to ensure
    + public_keys: list of public keys to attach to this user, ``home`` must be specified
    + group: the users primary group
    + delete_keys: whether to remove any keys not specified in ``public_keys``

    Public keys:
        These can be provided as strings containing the public key or as a path to
        a public key file which ``pyinfra`` will read.

    **Examples:**

    .. code:: python

        server.user_authorized_keys(
            name="Ensure user has a public key",
            user="kevin",
            public_keys=["ed25519..."],
        )
    """
    operargs = {
        "user": repr(user),
        "public_keys": repr(public_keys),
        "group": repr(group),
        "delete_keys": repr(delete_keys),
        "authorized_key_directory": repr(authorized_key_directory),
        "authorized_key_filename": repr(authorized_key_filename),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.user_authorized_keys", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def user(
    user,
    present=True,
    home=None,
    shell=None,
    group=None,
    groups=None,
    public_keys=None,
    delete_keys=False,
    ensure_home=True,
    create_home=False,
    system=False,
    uid=None,
    comment=None,
    add_deploy_dir=True,
    unique=True,
):
    """
    Add/remove/update system users & their ssh `authorized_keys`.

    + user: name of the user to ensure
    + present: whether this user should exist
    + home: the users home directory
    + shell: the users shell
    + group: the users primary group
    + groups: the users secondary groups
    + public_keys: list of public keys to attach to this user, ``home`` must be specified
    + delete_keys: whether to remove any keys not specified in ``public_keys``
    + ensure_home: whether to ensure the ``home`` directory exists
    + create_home: whether to new user create home directories from the system skeleton
    + system: whether to create a system account
    + uid: use a specific userid number
    + comment: the user GECOS comment
    + add_deploy_dir: any public_key filenames are relative to the deploy directory
    + unique: prevent creating users with duplicate UID

    Home directory:
        When ``ensure_home`` or ``public_keys`` are provided, ``home`` defaults to
        ``/home/{name}``. When ``create_home`` is ``True`` any newly created users
        will be created with the ``-m`` flag to build a new home directory from the
        systems skeleton directory.

    Public keys:
        These can be provided as strings containing the public key or as a path to
        a public key file which ``pyinfra`` will read.

    **Examples:**

    .. code:: python

        server.user(
            name="Ensure user is removed",
            user="kevin",
            present=False,
        )

        server.user(
            name="Ensure myweb user exists",
            user="myweb",
            shell="/bin/bash",
        )

        # multiple users
        for user in ["kevin", "bob"]:
            server.user(
                name=f"Ensure user {user} is removed",
                user=user,
                present=False,
            )
    """
    operargs = {
        "user": repr(user),
        "present": repr(present),
        "home": repr(home),
        "shell": repr(shell),
        "group": repr(group),
        "groups": repr(groups),
        "public_keys": repr(public_keys),
        "delete_keys": repr(delete_keys),
        "ensure_home": repr(ensure_home),
        "create_home": repr(create_home),
        "system": repr(system),
        "uid": repr(uid),
        "comment": repr(comment),
        "add_deploy_dir": repr(add_deploy_dir),
        "unique": repr(unique),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.user", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def locale(locale, present=True):
    """
    Enable/Disable locale.

    + locale: name of the locale to enable/disable
    + present: whether this locale should be present or not

    **Examples:**

    .. code:: python

        server.locale(
            name="Ensure en_GB.UTF-8 locale is not present",
            locale="en_GB.UTF-8",
            present=False,
        )

        server.locale(
            name="Ensure en_GB.UTF-8 locale is present",
            locale="en_GB.UTF-8",
        )
    """
    operargs = {
        "locale": repr(locale),
        "present": repr(present),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.locale", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def remove_any_askpass_file(state, host):
    operargs = {
        "state": repr(state),
        "host": repr(host),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server",
        "server.remove_any_askpass_file",
        operargs,
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def wait_and_reconnect(state, host):
    operargs = {
        "state": repr(state),
        "host": repr(host),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.wait_and_reconnect", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def partition(predicate, iterable):
    operargs = {
        "predicate": repr(predicate),
        "iterable": repr(iterable),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.partition", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def comma_sep(value):
    operargs = {
        "value": repr(value),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server", "server.comma_sep", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def read_any_pub_key_file(key):
    operargs = {
        "key": repr(key),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import server",
        "server.read_any_pub_key_file",
        operargs,
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## xbps


Manage XBPS packages and repositories. Note that XBPS package names are case-sensitive.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def upgrade():
    """
    Upgrades all XBPS packages.
    """
    operargs = {}

    result = _run_pyinfra(
        "from pyinfra.operations import xbps", "xbps.upgrade", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def update():
    """
    Update XBPS repositories.
    """
    operargs = {}

    result = _run_pyinfra(
        "from pyinfra.operations import xbps", "xbps.update", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def packages(packages=None, present=True, update=False, upgrade=False):
    """
    Install/remove/update XBPS packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + update: run ``xbps-install -S`` before installing packages
    + upgrade: run ``xbps-install -y -u`` before installing packages

    **Example:**

    .. code:: python

        xbps.packages(
            name="Install Vim and Vim Pager",
            packages=["vimpager", "vim"],
        )
    """
    operargs = {
        "packages": repr(packages),
        "present": repr(present),
        "update": repr(update),
        "upgrade": repr(upgrade),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import xbps", "xbps.packages", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## SSH (Secure SHell) tasks

This module provides tasks for using SSH to copy files to/from remote machines and running commands.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def keyscan(hostname, force=False, port=22):
    """
    Check/add hosts to the ``~/.ssh/known_hosts`` file.

    + hostname: hostname that should have a key in ``known_hosts``
    + force: if the key already exists, remove and rescan

    **Example:**

    .. code:: python

        ssh.keyscan(
            name="Set add server two to known_hosts on one",
            hostname="two.example.com",
        )
    """
    operargs = {
        "hostname": repr(hostname),
        "force": repr(force),
        "port": repr(port),
    }

    result = _run_pyinfra("from pyinfra.operations import ssh", "ssh.keyscan", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def command(hostname, command, user=None, port=22):
    """
    Execute commands on other servers over SSH.

    + hostname: the hostname to connect to
    + command: the command to execute
    + user: connect with this user
    + port: connect to this port

    **Example:**

    .. code:: python

        ssh.command(
            name="Create file by running echo from host one to host two",
            hostname="two.example.com",
            command="echo 'one was here' > /tmp/one.txt",
            user="vagrant",
        )
    """
    operargs = {
        "hostname": repr(hostname),
        "command": repr(command),
        "user": repr(user),
        "port": repr(port),
    }

    result = _run_pyinfra("from pyinfra.operations import ssh", "ssh.command", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def upload(
    hostname,
    filename,
    remote_filename=None,
    port=22,
    user=None,
    use_remote_sudo=False,
    ssh_keyscan=False,
):
    """
    Upload files to other servers using ``scp``.

    + hostname: hostname to upload to
    + filename: file to upload
    + remote_filename: where to upload the file to (defaults to ``filename``)
    + port: connect to this port
    + user: connect with this user
    + use_remote_sudo: upload to a temporary location and move using sudo
    + ssh_keyscan: execute ``ssh.keyscan`` before uploading the file
    """
    operargs = {
        "hostname": repr(hostname),
        "filename": repr(filename),
        "remote_filename": repr(remote_filename),
        "port": repr(port),
        "user": repr(user),
        "use_remote_sudo": repr(use_remote_sudo),
        "ssh_keyscan": repr(ssh_keyscan),
    }

    result = _run_pyinfra("from pyinfra.operations import ssh", "ssh.upload", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def download(
    hostname,
    filename,
    local_filename=None,
    force=False,
    port=22,
    user=None,
    ssh_keyscan=False,
):
    """
    Download files from other servers using ``scp``.

    + hostname: hostname to upload to
    + filename: file to download
    + local_filename: where to download the file to (defaults to ``filename``)
    + force: always download the file, even if present locally
    + port: connect to this port
    + user: connect with this user
    + ssh_keyscan: execute ``ssh.keyscan`` before uploading the file
    """
    operargs = {
        "hostname": repr(hostname),
        "filename": repr(filename),
        "local_filename": repr(local_filename),
        "force": repr(force),
        "port": repr(port),
        "user": repr(user),
        "ssh_keyscan": repr(ssh_keyscan),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import ssh", "ssh.download", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## vzctl


Manage OpenVZ containers with ``vzctl``.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def start(ctid, force=False):
    """
    Start OpenVZ containers.

    + ctid: CTID of the container to start
    + force: whether to force container start
    """
    operargs = {
        "ctid": repr(ctid),
        "force": repr(force),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import vzctl", "vzctl.start", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def stop(ctid):
    """
    Stop OpenVZ containers.

    + ctid: CTID of the container to stop
    """
    operargs = {
        "ctid": repr(ctid),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import vzctl", "vzctl.stop", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def restart(ctid, force=False):
    """
    Restart OpenVZ containers.

    + ctid: CTID of the container to restart
    + force: whether to force container start
    """
    operargs = {
        "ctid": repr(ctid),
        "force": repr(force),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import vzctl", "vzctl.restart", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def mount(ctid):
    """
    Mount OpenVZ container filesystems.

    + ctid: CTID of the container to mount
    """
    operargs = {
        "ctid": repr(ctid),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import vzctl", "vzctl.mount", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def unmount(ctid):
    """
    Unmount OpenVZ container filesystems.

    + ctid: CTID of the container to unmount
    """
    operargs = {
        "ctid": repr(ctid),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import vzctl", "vzctl.unmount", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def delete(ctid):
    """
    Delete OpenVZ containers.

    + ctid: CTID of the container to delete
    """
    operargs = {
        "ctid": repr(ctid),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import vzctl", "vzctl.delete", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def create(ctid, template=None):
    """
    Create OpenVZ containers.

    + ctid: CTID of the container to create
    """
    operargs = {
        "ctid": repr(ctid),
        "template": repr(template),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import vzctl", "vzctl.create", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def set(ctid, save=True, **settings):
    """
    Set OpenVZ container details.

    + ctid: CTID of the container to set
    + save: whether to save the changes
    + settings: settings/arguments to apply to the container

    Settings/arguments:
        these are mapped directly to ``vztctl`` arguments, eg
        ``hostname='my-host.net'`` becomes ``--hostname my-host.net``.
    """
    operargs = {
        "ctid": repr(ctid),
        "save": repr(save),
    }

    result = _run_pyinfra("from pyinfra.operations import vzctl", "vzctl.set", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## lxd tasks

This module provides tasks for managing lxd containers.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def get_container_named(name, containers):
    operargs = {
        "name": repr(name),
        "containers": repr(containers),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import lxd", "lxd.get_container_named", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def container(id, present=True, image="ubuntu:16.04"):
    """
    Add/remove LXD containers.

    Note: does not check if an existing container is based on the specified
    image.

    + id: name/identifier for the container
    + image: image to base the container on
    + present: whether the container should be present or absent

    **Example:**

    .. code:: python

        lxd.container(
            name="Add an ubuntu container",
            id="ubuntu19",
            image="ubuntu:19.10",
        )
    """
    operargs = {
        "id": repr(id),
        "present": repr(present),
        "image": repr(image),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import lxd", "lxd.container", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
A Python-like declarative IT automation tool.

uPlaybook2 takes ideas from Ansible and Cookiecutter and gives it a
Python syntax rather than YAML. The desired state of the system is
specified via this "playbook" and associated templates and data. You
can then run this playbook, providing additional arguments for this run.

More documentation is available at: https://github.com/linsomniac/uplaybook2
"""

from . import internals
from . import fs
from . import core

up_context = internals.up_context
ARGS = up_context.context["ARGS"]



#!/usr/bin/env python3

"""
Filesystem Related Tasks

Tasks to manipulate the filesystem: create/delete files and directories, manage
permissions, etc...

There are individual tasks that are named using UNIX conventions.  There are also
higher-level tasks: `builder()` and `fs()`, which are controlled by the `action` argument.

## fs.builder -- The file-system power-house

`fs.builder()` is a power-house for setting up your file-system, as it takes a list of
`core.Item`s to streamline the creation large swaths of your file-system.

For example, you could:

    - Write an apt.sources.list file, and trigger an apt update if written.
    - Create a var directory.
    - Template a systemd file and trigger a daemon-reload if written.
    - Create a /usr/local/src directory.
    - Template a docker-compose file into it.
    - Trigger a `docker compose` run if updated.

Here is an example:

```python
def semaphore_compose():
    with fs.cd(path="/usr/local/src/ansible-semaphore"):
        core.run(cmd="docker compose up -d")

fs.builder([
    core.Item(path="/usr/local/src/ansible-semaphore", state="directory"),
    core.item(path="/usr/local/src/ansible-semaphore/docker-compose.yml",
              notify=semaphore_compose),
    ])
```

## fs.fs -- the builder() back-end

`fs.fs()` is just like `fs.builder()`, except it does not take a list of `Item`s, it takes
the exploded arguments.  Mostly, it exists as a back-end for `fs.builder()`, though there
may be situations where it is useful, so it is left as exposed.

## Tasks
"""

from .internals import (
    Return,
    Failure,
    TemplateStr,
    RawStr,
    up_context,
    task,
    CallDepth,
    PasswordNeeded,
)
from .core import Item
from .fernetreader import FernetReader
from . import internals
from typing import Union, Optional, Callable, List
from types import SimpleNamespace
import symbolicmode
import os
import stat as stat_module
import random
import string
import hashlib
import shutil
from pathlib import Path


def _mode_from_arg(
    mode: Optional[Union[str, int]] = None,
    initial_mode: Optional[int] = None,
    is_directory: Optional[bool] = None,
) -> Optional[int]:
    """
    Helper function to convert a task `mode` argument into an int, if necessary.

    If `mode` is a string, it is converted using the symbolicmode module.
    If `mode` is None, it is kept as None (meaning no mode change to be done).
    If `mode` is an int, it is kept.

    Args:
        mode: Mode to convert, if it is a string.
        initial_mode: The existing mode of the file (used for +/-/X).
        is_directory: If the path to set the mode on is a directory (used for X).
    """
    if isinstance(mode, int) or mode is None:
        return mode

    assert isinstance(mode, str)

    mode_is_sym_str = isinstance(mode, str) and not set(mode).issubset("01234567")
    if mode_is_sym_str:
        extra_args = {}
        if is_directory is not None:
            extra_args["is_directory"] = is_directory
        if initial_mode is not None:
            extra_args["initial_mode"] = initial_mode

        return symbolicmode.symbolic_to_numeric_permissions(mode, **extra_args)

    return int(mode, 8)


@task
def chmod(
    path: str,
    mode: Optional[Union[str, int]] = None,
    is_directory: Optional[bool] = None,
) -> Return:
    """
    Change permissions of `path`.

    Args:
        path: Path to change (templateable).
        mode: Permissions of `path` (optional, templatable string or int).
        is_directory: Treat `path` as a directory, impacts "X".  If not specified
            `path` is examined to determine if it is a directory.
            (optional, bool).

    Examples:

    ```python
    fs.chmod(path="/tmp/foo", mode="a=rX,u+w")
    fs.chmod(path="/tmp/foo", mode=0o755)
    ```
    """
    if mode is None:
        return Return(
            changed=False, secret_args={"decrypt_password", "encrypt_password"}
        )

    path_stats = os.stat(path)
    current_mode = stat_module.S_IMODE(path_stats.st_mode)
    extra_args = {}
    if is_directory is not None:
        extra_args["is_directory"] = is_directory
    mode = _mode_from_arg(mode, initial_mode=current_mode, **extra_args)
    if current_mode != mode:
        assert isinstance(mode, int)
        os.chmod(path, mode)
        return Return(
            changed=True,
            secret_args={"decrypt_password", "encrypt_password"},
            extra_message=f"Changed permissions: {current_mode:o} -> {mode:o}",
        )

    return Return(changed=False, secret_args={"decrypt_password", "encrypt_password"})


@task
def chown(
    path: str,
    user: Optional[TemplateStr] = None,
    group: Optional[TemplateStr] = None,
) -> Return:
    """
    Change ownership/group of `path`.

    Args:
        path: Path to change (templateable).
        user: User to set on `path`. (optional, templatable).
        group: Group to set on `path`. (optional, templatable).

    Examples:

    ```python
    fs.chown(path="/tmp", owner="root")
    fs.chown(path="/tmp", group="wheel")
    fs.chown(path="/tmp", owner="nobody", group="nobody")
    ```
    """
    changed = False
    extra_messages = []

    before_stats = os.stat(path)
    shutil.chown(path, user=user, group=group)
    after_stats = os.stat(path)

    extra_messages = []
    if before_stats.st_uid != after_stats.st_uid:
        extra_messages.append(f"User changed from {before_stats.st_uid}")
    if before_stats.st_gid != after_stats.st_gid:
        extra_messages.append(f"Group changed from {before_stats.st_gid}")
    changed = len(extra_messages) != 0

    return Return(
        changed=changed,
        extra_message=", ".join(extra_messages) if extra_messages else None,
    )


@task
def cd(path: TemplateStr) -> Return:
    """
    Change working directory to `path`.

    Sets "extra.old_dir" on the return object to the directory before the `cd`
    is done.  Can also be used as a context manager and when the context is
    exited you are returned to the previous directory.

    Args:
        path: Directory to change into (templateable).

    Examples:

    ```python
    fs.cd(path="/tmp")

    #  As context manager:
    with fs.cd(path="/tmp"):
        #  creates /tmp/tempfile
        fs.mkfile("tempfile")
    #  now are back in previous directory
    ```
    """
    old_dir = os.getcwd()
    os.chdir(path)

    return Return(
        changed=False,
        extra=SimpleNamespace(old_dir=old_dir),
        context_manager=lambda: cd(old_dir),
    )


@task
def mkfile(
    path: TemplateStr,
    mode: Optional[Union[TemplateStr, int]] = None,
    contents: Optional[TemplateStr] = None,
) -> Return:
    """
    Create an empty file if it does not already exist.

    Args:
        path: Name of file to create (templateable).
        mode: Permissions of file (optional, templatable string or int).
              Atomically sets mode on creation.
        contents: If specified, ensure the contents of the file are `contents`.

    Examples:

    ```python
    fs.mkfile(path="/tmp/foo")
    fs.mkfile(path="/tmp/bar", mode="a=rX,u+w")
    fs.mkfile(path="/tmp/baz", mode=0o755)
    ```
    """
    if not os.path.exists(path):
        mode = _mode_from_arg(mode)
        mode_arg = {} if mode is None else {"mode": mode}
        fd = os.open(path, os.O_CREAT, **mode_arg)
        os.close(fd)

        if contents is not None:
            with open(path, "w") as fp:
                fp.write(contents)

        return Return(changed=True)

    changed = False
    if contents is not None:
        with open(path, "r") as fp:
            current_data = fp.read(len(contents) + 1)
        if current_data != contents:
            with open(path, "w") as fp:
                fp.write(contents)
            changed = True

    if mode is not None:
        with CallDepth():
            chmod(path, mode)

    return Return(changed=changed)


@task
def mkdir(
    path: TemplateStr,
    mode: Optional[Union[TemplateStr, int]] = None,
    parents: Optional[bool] = True,
) -> Return:
    """
    Create a directory.  Defaults to creating necessary parent directories.

    Args:
        path: Name of file to create (templateable).
        mode: Permissions of directory (optional, templatable string or int).
                Sets mode on creation.
        parents: Make parent directories if needed.  (optional, default=True)

    Examples:

    ```python

    fs.mkdir(path="/tmp/foo")
    fs.mkdir(path="/tmp/bar", mode="a=rX,u+w")
    fs.mkdir(path="/tmp/baz/qux", mode=0o755, parents=True)
    ```
    """
    if not os.path.exists(path):
        mode = _mode_from_arg(mode, is_directory=True)
        mode_arg = {} if mode is None else {"mode": mode}
        if parents:
            os.makedirs(path, **mode_arg)
        else:
            os.mkdir(path, **mode_arg)

        return Return(changed=True)

    with CallDepth():
        chmod(path, mode, is_directory=True)

    return Return(changed=False)


def _random_ext(i: int = 8) -> str:
    "Return a random string of length 'i'"
    return "".join(
        random.choices(
            string.ascii_lowercase + string.ascii_uppercase + string.digits, k=i
        )
    )


@task
def rm(
    path: TemplateStr,
    recursive: bool = False,
) -> Return:
    """
    Remove a file or recursively remove a directory.

    Args:
        path: Name of file/directory to remove. (templateable).
        recursive: If True, recursively remove directory and all contents of `path`.
            Otherwise only remove if `path` is a file.  (default: False)

    Examples:

    ```python
    fs.rm(path="/tmp/foo")
    fs.rm(path="/tmp/foo-dir", recursive=True)
    ```
    """

    if not os.path.exists(path):
        return Return(changed=False)

    if not recursive:
        try:
            os.remove(path)
        except OSError:
            return Return(
                changed=False,
                failure=True,
                raise_exc=Failure(
                    f"Path {path} is a directory, will not remove without `recursive` option"
                ),
            )
    else:
        shutil.rmtree(path)

    return Return(changed=True)


@task
def stat(
    path: TemplateStr,
    follow_symlinks: bool = True,
) -> Return:
    """
    Get information about `path`.

    Args:
        path: Path to stat.  (templateable).
        follow_symlinks: If True (default), the result will be on the destination of
            a symlink, if False the result will be about the symlink itself.
            (bool, default: True)

    Extra Data:

        perms: The permissions of `path` (st_mode & 0o777).
        st_mode: Full mode of `path` (permissions, object type).  You probably want the
            "perms" field if you just want the permissions of `path`.
        st_ino: Inode number.
        st_dev: ID of the device containing `path`.
        st_nlink: Number of hard links.
        st_uid: User ID of owner.
        st_gid: Group ID of owner.
        st_size: Total size in bytes.
        st_atime: The time of the last access of file data.
        st_mtime: The time of last modification of file data.
        st_ctime: The time of the last change of status/inode.
        S_ISBLK: Is `path` a block special device file?
        S_ISCHR: Is `path` a character special device file?
        S_ISDIR: Is `path` a directory?
        S_ISDOOR: Is `path` a door?
        S_ISFIFO: Is `path` a named pipe?
        S_ISLNK: Is `path` a symbolic link?
        S_ISPORT: Is `path` an event port?
        S_ISREG: Is `path` a regular file?
        S_ISSOCK: Is `path` a socket?
        S_ISWHT: Is `path` a whiteout?

    Examples:

    ```python
    stat = fs.stat(path="/tmp/foo")
    print(f"UID: {{stat.extra.st_uid}}")
    fs.stat(path="/tmp/foo", follow_symlinks=False)
    ```
    """

    s = os.stat(path, follow_symlinks=follow_symlinks)

    ret = SimpleNamespace(
        perms=stat.s_IMODE(s.st_mode),
        st_mode=s.st_mode,
        st_ino=s.st_ino,
        st_dev=s.st_dev,
        st_nlink=s.st_nlink,
        st_uid=s.st_uid,
        st_gid=s.st_gid,
        st_size=s.st_size,
        st_atime=s.st_atime,
        st_mtime=s.st_mtime,
        st_ctime=s.st_ctime,
        S_ISBLK=stat_module.S_ISBLK(s.st_mode),
        S_ISCHR=stat_module.S_ISCHR(s.st_mode),
        S_ISDIR=stat_module.S_ISDIR(s.st_mode),
        S_ISDOOR=stat_module.S_ISDOOR(s.st_mode),
        S_ISFIFO=stat_module.S_ISFIFO(s.st_mode),
        S_ISLNK=stat_module.S_ISLNK(s.st_mode),
        S_ISPORT=stat_module.S_ISPORT(s.st_mode),
        S_ISREG=stat_module.S_ISREG(s.st_mode),
        S_ISSOCK=stat_module.S_ISSOCK(s.st_mode),
        S_ISWHT=stat_module.S_ISWHT(s.st_mode),
    )

    return Return(changed=False, extra=ret)


@task
def mv(
    path: TemplateStr,
    src: TemplateStr,
) -> Return:
    """
    Rename `src` to `path`.  If `src` does not exist but `path` does,
    it is considered successful without change.  If neither exists,
    it is failed.

    Args:
        path: New name. (templateable).
        src: Old name. (templateable).

    Examples:

    ```python
    fs.mv(path="/tmp/foo", src="/tmp/bar")
    ```
    """

    if os.path.exists(src):
        shutil.move(src, path)
        return Return(changed=True)

    if os.path.exists(path):
        return Return(changed=False)

    return Return(
        changed=False,
        failure=True,
        raise_exc=Failure(f"No file to move: src={src} path={path}"),
    )


@task
def ln(
    path: TemplateStr,
    src: TemplateStr,
    symbolic: bool = False,
) -> Return:
    """
    Create a link from `src` to `path`.

    Args:
        path: Name of destination of link. (templateable).
        src: Name of location of source to create link from. (templateable).
        symbolic: If True, makes a symbolic link. (bool, default: False)

    Examples:

    ```python
    fs.ln(path="/tmp/foo", src="/tmp/bar")
    fs.ln(path="/tmp/foo", src="/tmp/bar", symbolic=True)
    ```
    """

    if os.path.isdir(path):
        path = os.path.join(path, os.path.basename(src))

    if symbolic:
        if os.path.islink(path) and os.readlink(path) == src:
            return Return(changed=False)
        if os.path.exists(path):
            os.remove(path)
        os.symlink(src, path)
    else:
        if os.path.exists(path):
            src_stat = os.stat(src)
            path_stat = os.stat(path)

            if (
                src_stat.st_dev == path_stat.st_dev
                and src_stat.st_ino == path_stat.st_ino
            ):
                return Return(changed=False)

            os.remove(path)

        os.link(src=src, path=path)

    return Return(changed=True)


@task
def write(
    path: TemplateStr,
    src: Optional[TemplateStr] = None,
    mode: Optional[Union[TemplateStr, int]] = None,
    encrypt_password: Optional[TemplateStr] = None,
    decrypt_password: Optional[TemplateStr] = None,
    template: bool = True,
    template_filenames: bool = True,
    recursive: bool = True,
) -> Return:
    """
    Write the `src` file(s) from the playbook to `path`.
    The `write()` task pulls the `src` file(s) from the playbook and writes them to
    `path`. This is typically used for templates or other files/data that are provided
    as a part of the playbook to write to the destination system (for example:
    configurations, scaffolding, etc).

    Optionally templating the contents in `src`.  It can also template file names when
    doing recursive operations.

    The `src` files are searched for using the `UP_FILES_PATH` search path.

    Args:
        path: Name of destination file. (templateable).
        src: Name of template to use as source (optional, templateable).
            Defaults to the basename of `path` + ".j2".
        mode: Permissions of directory (optional, templatable string or int).
            Sets mode on creation.
        template: If True, apply Jinja2 templating to the contents of `src`,
            otherwise copy verbatim.  (default: True)
        template_filenames: If True, filenames found during recursive copy are
            jinja2 template expanded. (default: True)
        recursive: If True and `src` is a directory, recursively copy it and
            everything below it to the `path`.  If `path` ends in a "/",
            the last component of `src` is created under `path`, otherwise
            the contents of `src` are written into `path`. (default: True)

    Examples:

    ```python
    fs.write(path="/tmp/foo")
    fs.write(src="bar-{{ fqdn }}.j2", path="/tmp/bar", template=False)
    ```
    """

    def _copy_file(
        src: Path,
        path: str,
        mode: Optional[Union[str, int]] = None,
        decrypt_password: Union[str, None] = None,
        encrypt_password: Union[str, None] = None,
    ) -> Optional[str]:
        """
        The workhorse of the copy function, copy one file.

        Returns:
            A string describing the change made, or None if no change made.
        """
        old_mode = None

        hash_before = None
        if os.path.exists(path):
            old_mode = stat_module.S_IMODE(os.stat(path).st_mode)
            with open(path, "rb") as fp_in:
                sha = hashlib.sha256()
                sha.update(fp_in.read())
                hash_before = sha.hexdigest()

        def UnknownPassword():
            raise PasswordNeeded(
                f"An encrypted file was found ({src}) but no decryption key was given"
            )

        to_decrypt = decrypt_password if decrypt_password else UnknownPassword

        encoding = "latin-1"
        with FernetReader(src.as_posix(), to_decrypt) as fp_in:
            data: bytes = fp_in.read()
            if template:
                data = (
                    up_context.jinja_env.from_string(data.decode(encoding))
                    .render(up_context.get_env())
                    .encode(encoding)
                )

        sha = hashlib.sha256()
        sha.update(data)
        hash_after = sha.hexdigest()

        if mode is not None:
            mode = _mode_from_arg(mode, initial_mode=old_mode)

        if hash_before == hash_after and (mode is None or mode == old_mode):
            return None
        if hash_before == hash_after and (mode is not None or mode != old_mode):
            return "Permissions"

        pathTmp = path + ".tmp." + _random_ext()
        mode_arg = {} if mode is None else {"mode": mode}
        fd = os.open(pathTmp, os.O_WRONLY | os.O_CREAT, **mode_arg)
        with os.fdopen(fd, "wb") as fp_out:
            fp_out.write(data)
        os.rename(pathTmp, path)

        return "Contents"

    src_name = src if src is not None else os.path.basename(path) + ".j2"
    src: Path = internals.find_file(src_name)

    if encrypt_password or decrypt_password:
        raise NotImplementedError("Crypto not implemented yet")

    changes_made = set()
    src_is_dir = stat_module.S_ISDIR(os.stat(src).st_mode)
    if recursive and src_is_dir:
        with CallDepth():
            for dirpath, dirnames, filenames in os.walk(src):
                path_dir = os.path.join(path, os.path.relpath(dirpath, src))

                if not template_filenames:
                    path_dir = RawStr(path_dir)
                r = mkdir(path=path_dir, mode=mode)
                if r.changed:
                    changes_made.add("Subdir")

                for filename in filenames:
                    src_file = RawStr(os.path.join(dirpath, filename))
                    path_file = os.path.join(path_dir, filename)
                    if not template_filenames:
                        path_file = RawStr(path_file)
                    r = write(src=src_file, path=path_file)
                    if r.changed:
                        changes_made.add("Subfile")
    else:
        change = _copy_file(src, path, mode)
        if change:
            changes_made.add(change)

    if not changes_made:
        return Return(
            changed=False, secret_args={"decrypt_password", "encrypt_password"}
        )
    return Return(
        changed=True,
        extra_message=", ".join(changes_made),
        secret_args={"decrypt_password", "encrypt_password"},
    )


@task
def fs(
    path: TemplateStr,
    action: Union[TemplateStr, str] = "template",
    src: Optional[TemplateStr] = None,
    mode: Optional[Union[TemplateStr, int]] = None,
    owner: Optional[TemplateStr] = None,
    group: Optional[TemplateStr] = None,
    notify: Optional[Callable] = None,
) -> Return:
    """
    This is a unified entry point for the `fs` module that can take paths and
    actions and call the other functions in the module based on the action.

    This is targeted for use with Items() loops, for easily populating or
    modifying many filesystem objects in compact declarations.

    Args:
        path: Name of destination filesystem object. (templateable).
        action: Type of `path` to build, can be: (optional, templatable, default="template")
            - "directory"
            - "template"
            - "exists"
            - "copy"
            - "absent"
            - "link"
            - "symlink"
        src: Name of template to use as source (optional, templateable).
            Defaults to the basename of `path` + ".j2".
        mode: Permissions of file (optional, templatable string or int).
        owner: Ownership to set on `path`. (optional, templatable).
        group: Group to set on `path`. (optional, templatable).
        notify:  Handler to notify of changes.
            (optional, Callable)

    Examples:

    ```python
    fs.fs(path="/tmp/foo")
    fs.fs(path="/tmp/bar", action="directory")
    for _ in [
            core.Item(path="/tmp/{{ modname }}", action="directory"),
            core.Item(path="/tmp/{{ modname }}/__init__.py"),
            ]:
        fs.fs()
    ```
    """

    with CallDepth():
        if action == "template":
            r = write(src=src, path=path)
        elif action == "copy":
            r = write(src=src, path=path, template=False)
        elif action == "directory":
            r = mkdir(path=path, mode=mode)
        elif action == "exists":
            r = mkfile(path=path, mode=mode)
        elif action == "link":
            r = ln(src=src, path=path)
        elif action == "symlink":
            r = ln(src=src, path=path, symbolic=True)
        elif action == "absent":
            r = rm(path=path)
        else:
            raise ValueError(f"Unknown action: {action}")

        if mode is not None:
            chmod(path, mode)
        if owner is not None or group is not None:
            chown(path, owner, group)

    if notify is not None:
        r = r.notify(notify)

    return Return(changed=r.changed)


@task
def builder(
    items: List[Item],
    defaults: Item = Item(),
) -> Return:
    """
    All-in-one filesystem builder.  This is the swiss-army knife of the fs module, given a list
    of all the filesystem changes that need to be accomplished, it will carry them out.

    This is targeted for use with Items() loops, for easily populating or
    modifying many filesystem objects in compact declarations.

    Args:
        items: A list of the file-system actions to take, `Item()` objects with attributes as below.
        defaults: An `Item()` specifying defaults that can be overridden by the specific
            items, for example if most owners should be "root", but a few need to be
            "nobody", specify `owner="root"` in the defaults.  (Optional)

    Item Attributes:
        Items can have these attributes (from `fs.fs`):

            path: Name of destination filesystem object. (templateable).
            action: Type of `path` to build, can be: (optional, templatable, default="template")
                - "directory"
                - "template"
                - "exists"
                - "copy"
                - "absent"
                - "link"
                - "symlink"
            src: Name of template to use as source (optional, templateable).
                Defaults to the basename of `path` + ".j2".
            mode: Permissions of file (optional, templatable string or int).
            owner: Ownership to set on `path`. (optional, templatable).
            group: Group to set on `path`. (optional, templatable).
            notify:  Handler to notify of changes.
                (optional, Callable)

    Examples:

    ```python
    fs.builder(items=[
        #  creates "/tmp/foo" from "foo.j2" as a template
        Item(path="/tmp/foo",

        #  makes the "/tmp/bar" directory
        Item(path="/tmp/bar", action="directory"),
        ]))


    fs.builder(defaults=Item(owner="root", group="root", mode="a=rX,u+w",
               items=[
                   Item(path="/tmp/{{ modname }}", action="directory"),
                   Item(path="/tmp/{{ modname }}/__init__.py"),
                   Item(path="/tmp/should-not-exist", action="absent"),
                   Item(path="/tmp/{{ modname }}/nobody-file", owner="nobody", group="nobody"),
                   Item(path="/tmp/{{ modname }}/site.conf", notify=restart_apache),
               ])
    ```
    """
    changed = False

    for item in items:
        args = defaults.copy()
        args.update(item)
        r = fs(**args)

        if r.changed:
            changed = True

    return Return(changed=changed)


@task
def exists(
    path: TemplateStr,
    ignore_failure: bool = True,
) -> object:
    """
    Does `path` exist?

    Args:
        path: File location to see if it exists. (templateable).
        ignore_failure: If True, do not treat file absence as a fatal failure.
             (optional, bool, default=True)

    Examples:

    ```python
    fs.exists(path="/tmp/foo")
    if fs.exists(path="/tmp/foo"):
        #  code for when file exists
    ```
    """
    if os.path.exists(path):
        return Return(changed=False)

    return Return(
        changed=False,
        failure=True,
        ignore_failure=ignore_failure,
        raise_exc=Failure(f"File does not exist: {path}")
        if not ignore_failure
        else None,
    )



#!/usr/bin/env python3

"""
## windows


The windows module handles misc windows operations.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def service(service, running=True, restart=False, suspend=False):
    """
    Stop/Start a Windows service.

    + service: name of the service to manage
    + running: whether the the service should be running or stopped
    + restart: whether the the service should be restarted
    + suspend: whether the the service should be suspended

    **Example:**

    .. code:: python

        windows.service(
            name="Stop the spooler service",
            service="service",
            running=False,
        )
    """
    operargs = {
        "service": repr(service),
        "running": repr(running),
        "restart": repr(restart),
        "suspend": repr(suspend),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import windows", "windows.service", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def reboot():
    """
    Restart the server.
    """
    operargs = {}

    result = _run_pyinfra(
        "from pyinfra.operations import windows", "windows.reboot", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## Apt package tasks

This module provides tasks for interacting with the apt package manager.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def packages(
    packages: Optional[List[TemplateStr]] = None,
    present: bool = True,
    latest: bool = False,
    update: bool = False,
    cache_time: Optional[int] = None,
    upgrade: bool = False,
    force: bool = False,
    no_recommends: bool = False,
    allow_downgrades: bool = False,
    #  need to figure out what these are, they aren't documented as to type
    # extra_install_args=None,
    # extra_uninstall_args=None,
) -> Return:
    """
    Manage packages with apt (install, uninstall, update)

    Package versions can be given, as with apt, "<package>=<version>".

    Args:
        packages: List of packages (templatable)
        present: Should packages be installed
        latest: Upgrade packages (if no specific version is given)
        update: Run apt update before installing packages
        cache_time: When used with update, cache for this many seconds
        upgrade: Run apt upgrade before installing packages
        force: Force package installs by passing –force-yes to apt
        no_recommends: Don’t install recommended packages
        allow_downgrades: Allow downgrading packages with version (–allow-downgrades)

    Not yet implemented:
        extra_install_args: Additional arguments to the apt install command
        extra_uninstall_args: additional arguments to the apt uninstall command

    Examples:

    ```python
    apt.packages(packages=["neovim"])
    apt.packages(packages=["neovim"], latest=True)
    ```
    """
    operargs = {
        "packages": repr(packages),
        "present": present,
        "latest": latest,
        "update": update,
        "upgrade": upgrade,
        "force": force,
        "no_recommends": no_recommends,
        "allow_downgrades": allow_downgrades,
        "cache_time": cache_time,
    }

    result = _run_pyinfra(
        "from pyinfra.operations import apt", "apt.packages", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def key(src=None, keyserver=None, keyid=None):
    """
    Add apt gpg keys with ``apt-key``.

    + src: filename or URL
    + keyserver: URL of keyserver to fetch key from
    + keyid: key ID or list of key IDs when using keyserver

    keyserver/id:
        These must be provided together.

    **Examples:**

    ```python
    # Note: If using URL, wget is assumed to be installed.
    apt.key(
        name="Add the Docker apt gpg key",
        src="https://download.docker.com/linux/ubuntu/gpg",
    )

    apt.key(
        name="Install VirtualBox key",
        src="https://www.virtualbox.org/download/oracle_vbox_2016.asc",
    )
    ```
    """
    operargs = {
        "src": repr(src),
        "keyserver": repr(keyserver),
        "keyid": repr(keyid),
    }

    result = _run_pyinfra("from pyinfra.operations import apt", "apt.key", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def repo(src, present=True, filename=None):
    """
    Add/remove apt repositories.

    + src: apt source string eg ``deb http://X hardy main``
    + present: whether the repo should exist on the system
    + filename: optional filename to use ``/etc/apt/sources.list.d/<filename>.list``. By
      default uses ``/etc/apt/sources.list``.

    **Example:**

    ```python
    apt.repo(
        name="Install VirtualBox repo",
        src="deb https://download.virtualbox.org/virtualbox/debian bionic contrib",
    )
    ```
    """
    operargs = {
        "src": repr(src),
        "present": repr(present),
        "filename": repr(filename),
    }

    result = _run_pyinfra("from pyinfra.operations import apt", "apt.repo", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def ppa(src, present=True):
    """
    Add/remove Ubuntu ppa repositories.

    + src: the PPA name (full ppa:user/repo format)
    + present: whether it should exist

    Note:
        requires ``apt-add-repository`` on the remote host

    **Example:**

    ```python
    # Note: Assumes software-properties-common is installed.
    apt.ppa(
        name="Add the Bitcoin ppa",
        src="ppa:bitcoin/bitcoin",
    )
    ```
    """
    operargs = {
        "src": repr(src),
        "present": repr(present),
    }

    result = _run_pyinfra("from pyinfra.operations import apt", "apt.ppa", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def deb(src, present=True, force=False):
    """
    Add/remove ``.deb`` file packages.

    + src: filename or URL of the ``.deb`` file
    + present: whether or not the package should exist on the system
    + force: whether to force the package install by passing `--force-yes` to apt

    Note:
        When installing, ``apt-get install -f`` will be run to install any unmet
        dependencies.

    URL sources with ``present=False``:
        If the ``.deb`` file isn't downloaded, pyinfra can't remove any existing
        package as the file won't exist until mid-deploy.

    **Example:**

    ```python
    # Note: Assumes wget is installed.
    apt.deb(
        name="Install Chrome via deb",
        src="https://dl.google.com/linux/direct/google-chrome-stable_current_amd64.deb",
    )
    ```
    """
    operargs = {
        "src": repr(src),
        "present": repr(present),
        "force": repr(force),
    }

    result = _run_pyinfra("from pyinfra.operations import apt", "apt.deb", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def update(cache_time=None):
    """
    Updates apt repositories.

    + cache_time: cache updates for this many seconds

    **Example:**

    ```python
    apt.update(
        name="Update apt repositories",
        cache_time=3600,
    )
    ```
    """
    operargs = {
        "cache_time": repr(cache_time),
    }

    result = _run_pyinfra("from pyinfra.operations import apt", "apt.update", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def upgrade(auto_remove=False):
    """
    Upgrades all apt packages.

    + autoremove: removes transitive dependencies that are no longer needed.

    **Example:**

    ```python
    # Upgrade all packages
    apt.upgrade(
        name="Upgrade apt packages",
    )

    # Upgrade all packages and remove unneeded transitive dependencies
    apt.upgrade(
        name="Upgrade apt packages and remove unneeded dependencies"
        auto_remove=True
    )
    ```
    """
    operargs = {
        "auto_remove": repr(auto_remove),
    }

    result = _run_pyinfra("from pyinfra.operations import apt", "apt.upgrade", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def dist_upgrade():
    """
    Updates all apt packages, employing dist-upgrade.

    **Example:**

    ```python
    apt.dist_upgrade(
        name="Upgrade apt packages using dist-upgrade",
    )
    ```
    """
    operargs = {}

    result = _run_pyinfra(
        "from pyinfra.operations import apt", "apt.dist_upgrade", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## puppet


"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def agent(server=None, port=None):
    """
    Run puppet agent

    + server: master server URL
    + port: puppet master port

    Note: Either 'USE_SUDO_LOGIN=True' or 'USE_SU_LOGIN=True'
    for puppet.agent() as `puppet` is added to the path in
    the .bash_profile.

    **Example:**

    .. code:: python

        puppet.agent()

        # We also expect a return code of:
        # 0=no changes or 2=changes applied
        puppet.agent(
            name="Run the puppet agent",
            success_exit_codes=[0, 2],
        )
    """
    operargs = {
        "server": repr(server),
        "port": repr(port),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import puppet", "puppet.agent", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## gem


Manage Ruby gem packages. (see https://rubygems.org/ )
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def packages(packages=None, present=True, latest=False):
    """
    Add/remove/update gem packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version

    Versions:
        Package versions can be pinned like gem: ``<pkg>:<version>``.

    **Example:**

    .. code:: python

        # Note: Assumes that 'gem' is installed.
        gem.packages(
            name="Install rspec",
            packages=["rspec"],
        )
    """
    operargs = {
        "packages": repr(packages),
        "present": repr(present),
        "latest": repr(latest),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import gem", "gem.packages", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## openrc


Manage OpenRC init services.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def service(
    service,
    running=True,
    restarted=False,
    reloaded=False,
    command=None,
    enabled=None,
    runlevel="default",
):
    """
    Manage the state of OpenRC services.

    + service: name of the service to manage
    + running: whether the service should be running
    + restarted: whether the service should be restarted
    + reloaded: whether the service should be reloaded
    + command: custom command to pass like: ``rc-service <service> <command>``
    + enabled: whether this service should be enabled/disabled on boot
    + runlevel: runlevel to manage services for
    """
    operargs = {
        "service": repr(service),
        "running": repr(running),
        "restarted": repr(restarted),
        "reloaded": repr(reloaded),
        "command": repr(command),
        "enabled": repr(enabled),
        "runlevel": repr(runlevel),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import openrc", "openrc.service", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## Postgresql Database tasks

This module provides tasks for working with PostgreSQL databases.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def _translate_legacy_args(func):
    operargs = {
        "func": repr(func),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import postgresql",
        "postgresql._translate_legacy_args",
        operargs,
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def sql(
    sql,
    database=None,
    psql_user=None,
    psql_password=None,
    psql_host=None,
    psql_port=None,
):
    """
    Execute arbitrary SQL against PostgreSQL.

    + sql: SQL command(s) to execute
    + database: optional database to execute against
    + psql_*: global module arguments, see above
    """
    operargs = {
        "sql": repr(sql),
        "database": repr(database),
        "psql_user": repr(psql_user),
        "psql_password": repr(psql_password),
        "psql_host": repr(psql_host),
        "psql_port": repr(psql_port),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import postgresql", "postgresql.sql", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def role(
    role,
    present=True,
    password=None,
    login=True,
    superuser=False,
    inherit=False,
    createdb=False,
    createrole=False,
    replication=False,
    connection_limit=None,
    psql_user=None,
    psql_password=None,
    psql_host=None,
    psql_port=None,
):
    """
    Add/remove PostgreSQL roles.

    + role: name of the role
    + present: whether the role should be present or absent
    + password: the password for the role
    + login: whether the role can login
    + superuser: whether role will be a superuser
    + inherit: whether the role inherits from other roles
    + createdb: whether the role is allowed to create databases
    + createrole: whether the role is allowed to create new roles
    + replication: whether this role is allowed to replicate
    + connection_limit: the connection limit for the role
    + psql_*: global module arguments, see above

    Updates:
        pyinfra will not attempt to change existing roles - it will either
        create or drop roles, but not alter them (if the role exists this
        operation will make no changes).

    **Example:**

    .. code:: python

        postgresql.role(
            name="Create the pyinfra PostgreSQL role",
            role="pyinfra",
            password="somepassword",
            superuser=True,
            login=True,
            sudo_user="postgres",
        )
    """
    operargs = {
        "role": repr(role),
        "present": repr(present),
        "password": repr(password),
        "login": repr(login),
        "superuser": repr(superuser),
        "inherit": repr(inherit),
        "createdb": repr(createdb),
        "createrole": repr(createrole),
        "replication": repr(replication),
        "connection_limit": repr(connection_limit),
        "psql_user": repr(psql_user),
        "psql_password": repr(psql_password),
        "psql_host": repr(psql_host),
        "psql_port": repr(psql_port),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import postgresql", "postgresql.role", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def database(
    database,
    present=True,
    owner=None,
    template=None,
    encoding=None,
    lc_collate=None,
    lc_ctype=None,
    tablespace=None,
    connection_limit=None,
    psql_user=None,
    psql_password=None,
    psql_host=None,
    psql_port=None,
):
    """
    Add/remove PostgreSQL databases.

    + name: name of the database
    + present: whether the database should exist or not
    + owner: the PostgreSQL role that owns the database
    + template: name of the PostgreSQL template to use
    + encoding: encoding of the database
    + lc_collate: lc_collate of the database
    + lc_ctype: lc_ctype of the database
    + tablespace: the tablespace to use for the template
    + connection_limit: the connection limit to apply to the database
    + psql_*: global module arguments, see above

    Updates:
        pyinfra will not attempt to change existing databases - it will either
        create or drop databases, but not alter them (if the db exists this
        operation will make no changes).

    **Example:**

    .. code:: python

        postgresql.database(
            name="Create the pyinfra_stuff database",
            database="pyinfra_stuff",
            owner="pyinfra",
            encoding="UTF8",
            sudo_user="postgres",
        )
    """
    operargs = {
        "database": repr(database),
        "present": repr(present),
        "owner": repr(owner),
        "template": repr(template),
        "encoding": repr(encoding),
        "lc_collate": repr(lc_collate),
        "lc_ctype": repr(lc_ctype),
        "tablespace": repr(tablespace),
        "connection_limit": repr(connection_limit),
        "psql_user": repr(psql_user),
        "psql_password": repr(psql_password),
        "psql_host": repr(psql_host),
        "psql_port": repr(psql_port),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import postgresql", "postgresql.database", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def dump(
    dest,
    database=None,
    psql_user=None,
    psql_password=None,
    psql_host=None,
    psql_port=None,
):
    """
    Dump a PostgreSQL database into a ``.sql`` file. Requires ``pg_dump``.

    + dest: name of the file to dump the SQL to
    + database: name of the database to dump
    + psql_*: global module arguments, see above

    **Example:**

    .. code:: python

        postgresql.dump(
            name="Dump the pyinfra_stuff database",
            dest="/tmp/pyinfra_stuff.dump",
            database="pyinfra_stuff",
            sudo_user="postgres",
        )
    """
    operargs = {
        "dest": repr(dest),
        "database": repr(database),
        "psql_user": repr(psql_user),
        "psql_password": repr(psql_password),
        "psql_host": repr(psql_host),
        "psql_port": repr(psql_port),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import postgresql", "postgresql.dump", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def load(
    src,
    database=None,
    psql_user=None,
    psql_password=None,
    psql_host=None,
    psql_port=None,
):
    """
    Load ``.sql`` file into a database.

    + src: the filename to read from
    + database: name of the database to import into
    + psql_*: global module arguments, see above

    **Example:**

    .. code:: python

        postgresql.load(
            name="Import the pyinfra_stuff dump into pyinfra_stuff_copy",
            src="/tmp/pyinfra_stuff.dump",
            database="pyinfra_stuff_copy",
            sudo_user="postgres",
        )
    """
    operargs = {
        "src": repr(src),
        "database": repr(database),
        "psql_user": repr(psql_user),
        "psql_password": repr(psql_password),
        "psql_host": repr(psql_host),
        "psql_port": repr(psql_port),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import postgresql", "postgresql.load", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def decorated_func(*args, **kwargs):
    operargs = {}

    result = _run_pyinfra(
        "from pyinfra.operations import postgresql",
        "postgresql.decorated_func",
        operargs,
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## pkg


Manage BSD packages and repositories. Note that BSD package names are case-sensitive.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def packages(packages=None, present=True, pkg_path=None):
    """
    Install/remove/update pkg packages. This will use ``pkg ...`` where available
    (FreeBSD) and the ``pkg_*`` variants elsewhere.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + pkg_path: the PKG_PATH environment variable to set

    pkg_path:
        By default this is autogenerated as follows (tested/working for OpenBSD):
        ``http://ftp.<OS>.org/pub/<OS>/<VERSION>/packages/<ARCH>/``. Note that OpenBSD's
        official mirrors only hold the latest two versions packages.

        NetBSD/FreeBSD helpfully use their own directory structures, so the default won't
        work.

    **Example:**

    .. code:: python

        pkg.packages(
            name="Install Vim and Vim Addon Manager",
            packages=["vim-addon-manager", "vim"],
        )
    """
    operargs = {
        "packages": repr(packages),
        "present": repr(present),
        "pkg_path": repr(pkg_path),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import pkg", "pkg.packages", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)
